/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the classes for creating and transferring packets
 * between the elements of the network, used to implement the OWAMP protocol.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 2.1
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */


import java.net.*;
import java.io.*;
import java.util.Calendar;


public class Packets extends OWAMP_Base{

  // Class Variables

  static final long _2POW31 = ((long)2147483647 + 1);  // 2 pow 31 = 2147483648

  // Constructor

  public Packets(int length){
    buffer = new byte[length];
  }

  public Packets(){
  }


  // Instance Variables

  protected byte[] buffer;                        // buffer is an array which will have the information of
                                               // a packet.

  // Instance Methods

  /**
   * Most (all?) modern systems use 2's complement representation for negative numbers. In this
   * representation the distinction between signed and unsigned types really comes down to how the
   * high order bit is interpreted. For a Java int quantity, there are 32 bits. The low order 31
   * bits represent a value between 0 and 2^31 - 1. The top bit represents -2^31 if it is set, and
   * zero if it is not. The value of the quantity is the sum of the values represented by the low
   * order 31 bits and the high order bit. For unsigned int quantities, the high order bit represents
   * 2^31 if it is set, and 0 if it is not.
   * This method converts the value of an unsigned int variable to a variable of type signed long. We use
   * a variable of type long (64 bits) because it can keep any unsigned int values without overflow.
   * @param val int - unsigned int value to convert to signed long.
   * @return long - signed long value returned.
   */

  public static long unsignedIntToLong(int val) {

    if((val&0x80000000) == 0x80000000){      // For unsigned int quantities, the high order bit represents
                                             // 2^31 if it is set, and 0 if it is not.
      long aux = (long)( val & 0x7FFFFFFF ); // The value of the quantity is the sum of the values
                                             // represented by the low order 31 bits and the high order bit.
      return ((long)(aux + _2POW31));

    }

    return ((long)val);

  }


  /**
   * This method converts an signed long value to an unsigned int value.
   * @param val long - signed long value
   * @return int - int value returned.
   */

  public int longToUnsignedInt(long val){
    int a = 0;

    if(val > (_2POW31 -1)){
      a = (int)(val - _2POW31);
      return((a | (0x1 << 31)));
    }else{
      return((int)val);
    }
  }


  /**
   * This method is used to copy the value of a int variable to the buffer beginning in its position 'pos'.
   * @param val int - int value to be copied to the buffer.
   * @param pos int - the first index of array buffer where the int value should be copied to.
   */

  public void Int_To_Buffer(int val,int pos){ //
    valToBuffer(val,4,pos);
  }


  /**
   * This method is used to copy the value of a short variable to the buffer beginning in its position 'pos'.
   * @param val short - short value to be copied to the buffer.
   * @param pos int - the first index of array buffer where the short value should be copied to.
   */

  public void Short_To_Buffer(short val,int pos){
    valToBuffer(val,2,pos);
  }


  /**
   * This method is used to copy the values of a byte array to the buffer beginning in its position 'pos'.
   * @param buf byte[] - byte array to be copied to the buffer.
   * @param pos int - the first index of array buffer where the byte array should be copied to.
   */

  public void Byte_To_Buffer(byte[] buf,int pos){
    for(int i=0;i<buf.length;i++)
      buffer[pos+i] = buf[i];
  }


  /**
   * This method is used to copy the value of a long variable to the buffer beginning in its position 'pos'.
   * @param val long - long value to be copied to the buffer.
   * @param valLength int - The parameter valLength represents the length of the variable represented as a
   * long value. It can be an int (valLength = 4) or a short (valLength = 2) value.
   * @param pos int - the first index of array buffer where the long value should be copied to.
   */

  public void valToBuffer(long val, int valLength, int pos){
    for (int iInd = 0; iInd < valLength; ++iInd){ // for each byte of the variable val.
      buffer[pos+valLength-1-iInd] = (byte) ((val >> (iInd*8)) & 0x00FF);
    }
  }


  /**
   * This method is used to convert a byte value to an int value.
   * @param b byte - byte value to be converted.
   * @return int - int value returned.
   */

  public static int unsignedByteToInt(byte b) {
   return((int) b & 0xFF);
  }


  /**
   * This method is used to convert an unsigned byte value to a short value.
   * @param b byte - byte value to be converted.
   * @return short - short value returned.
   */

  public static short unsignedByteToShort(byte b) {
    if ( (b & 0x80) == 0x80)return (short) (128 + (b & 0x7f));
    else return (short) b;
  }


  /**
   * This method converts an array of 4 bytes, representing a signed int value, to a signed int value.
   * @param array byte[] - byte which elements represents a signed int value.
   * @param pos int - index of first byte of the four bytes that represent a signed int value.
   * @return int - int value returned.
   */

  public static int getIntFromByteArray(byte[] array,int pos) {
    int i = 0;

    i += unsignedByteToInt(array[pos++]) << 24;
    i += unsignedByteToInt(array[pos++]) << 16;
    i += unsignedByteToInt(array[pos++]) << 8;
    i += unsignedByteToInt(array[pos++]) << 0;

    return i;
  }


  /**
   * This method converts an array of 4 bytes, representing an unsigned int value, to a long value.
   * @param array byte[] - byte which elements represents an unsigned int value.
   * @param pos int - index of first byte of the four bytes that represent an unsigned int value.
   * @return long - long value returned.
   */

  public static long getUnsignedIntFromByteArray(byte[] array,int pos) {

    if((array[pos]&0x80) == 0x80){
      long i = 0;

      i += ((array[pos++])&(0x7f)) << 24;
      i += unsignedByteToInt(array[pos++]) << 16;
      i += unsignedByteToInt(array[pos++]) << 8;
      i += unsignedByteToInt(array[pos++]) << 0;

      return (i + 2147483647 + 1);
    }

    int i=0;

    i += unsignedByteToInt(array[pos++]) << 24;
    i += unsignedByteToInt(array[pos++]) << 16;
    i += unsignedByteToInt(array[pos++]) << 8;
    i += unsignedByteToInt(array[pos++]) << 0;

    return ((long)i);

  }


  /**
   * This method converts an array of 2 bytes, representing a signed short value, to a signed short value.
   * @param array byte[] - byte which elements represents a signed short value.
   * @param pos int - index of first byte of the four bytes that represent a signed short value.
   * @return int - short value returned.
   */

  public short getShortFromByteArray(byte[] array,int pos) {
     short i = 0;

     i += unsignedByteToInt(array[pos++]) << 8;
     i += unsignedByteToInt(array[pos++]) << 0;

     return i;
  }


  /**
   * This method is used to transfer the first buffer_length bytes of variable buffer to the TCP socket
   * theSocket. Using this method we can send the packets used in OWAMP Control, through the OWAMP Control
   * TCP connection.
   * @param theSocket Socket - TCP socket to be used.
   * @param buffer_length int - number of first bytes of variable buffer to be transferred.
   * @return boolean - this method returns true if every thing goes ok, otherwise it returns false.
   */

  public boolean sendTCP(Socket theSocket, int buffer_length){

    try{
      OutputStream theOutputStream = theSocket.getOutputStream();
      theOutputStream.write(buffer,0,buffer_length);
      theOutputStream.flush();
      return true;
    }
    catch(IOException e){
      System.err.println(e);
      return false;
    }
  }


  /**
   * This method is used to send the OWAMP test packet over UDP to a host machine.
   * @param ia InetAddress - address of the machine waiting to receive the test packets.
   * @param ds DatagramSocket - the DatagramSocket to be used to send the packet.
   * @param port int - port in which this machine is waiting test packets.
   * @return boolean - this method returns true if every thing goes ok, otherwise it returns false.
   */

  public boolean sendUDP(InetAddress ia, DatagramSocket ds, int port){
    try{

      DatagramPacket dp = new DatagramPacket(buffer,buffer.length,ia,port);
      ds.send(dp);

      return true;
    }catch(IOException e){
      System.err.println("Error: "+e);
      return false;
    }catch (Exception e) {
      System.err.println("Error: "+e);
      return false;
    }

  }


  /**
   * This Method test the IntegrityZeroPadding packet. Integrity Zero Padding should be all zero.
   * @param _buffer byte[] - the byte array with a IntegrityZeroPadding field to be tested.
   * @param pos short - is the index, in array buffer, where the Integrity Zero Padding begins.
   * @param length short - is the length of the Integrity Zero Padding.
   * @return boolean - This method returns true if everything goes ok, otherwise it returns false.
   */

  public boolean testIntegrityZeroPadding(byte[] _buffer, short pos, short length){
    for(int i=0;i<length;i++)   //Integrity Zero Padding should be all zero
      if(_buffer[pos+i] != 0){
        System.err.println("Error: Integrity Zero Padding should be all zero");
        return false;
      }
    return true;
  }


  /**
   * This method is used to find the index in arrayOfdataRS of a session that has the same sid as the
   * sid value given as parameter.
   * @param sid byte[] - a byte array with the sid value to be used.
   * @param arrayOfdataRS RequestSession[] - array of all Request Sessions objects to be used.
   * @param arrayOfSocketToS_R Socket[] - array of socket variables, representing socket connections to
   * Sender or Receiver (If the socket connection to Sender/Receiver is closed the respective session has
   * finished).
   * @param numRS int - number of Request Session made.
   * @return int - this method returns the index of the session if one of the elements of arrayOfdataRS has
   * the same sid, otherwise it returns -1.
   */

  public int findSession(byte[] sid, RequestSession[] arrayOfdataRS, Socket[] arrayOfSocketToS_R, int numRS){

    for(int i = 0, j = 0; i<numRS; i++){
      if(!arrayOfSocketToS_R[i].isClosed()){
        for (j = 0; j < sid.length; j++){
          if (sid[j] != arrayOfdataRS[i].sid[j])
            break;
        }
        if(j==sid.length)
          return i;
      }
    }
    return -1;
  }


}


/******************************************************************************/
/*          Class definition for each packet of OWAMP Protocol                */
/******************************************************************************/


                         //-------------------//
                         //  OWAMP - Control  //
                         //-------------------//



//------------------//
// Connection Setup //
//------------------//



// This class is used to implement the Connection Setup Server Greeting

class ConnectionSetupServerGreeting extends Packets{

  static final int PACKET_LENGTH = 32;

  public ConnectionSetupServerGreeting(int _modes, byte _challenge[]){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    for (int c=0;c<unused.length;c++)
      unused[c]=0;   // Unused should be zero
    modes = _modes;
    System.arraycopy(_challenge,0,challenge,0,_challenge.length);
  }

  public ConnectionSetupServerGreeting(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  byte unused[] = new byte [12];
  int modes;                        // 1 for unauthenticated, 2 for authenticated and 4 for encrypted.
  byte challenge[] = new byte [16]; // random sequence of octets generated by the server.

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_ConnectionSetupServerGreeting(){
    for (int i=0;i<unused.length;i++)
      System.out.println(" Unused "+i+" "+unused[i]);
    System.out.println(" Modes "+modes);
    for (int i=0;i<challenge.length;i++)
      System.out.println(" challenge "+i+" "+challenge[i]);
  }

}

class ConnectionSetupServerGreetingToBeSend extends ConnectionSetupServerGreeting{

  public ConnectionSetupServerGreetingToBeSend(int _modes, byte _challenge[]){
    super(_modes, _challenge);    // Constructor of main class
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    Byte_To_Buffer(unused,0);
    Byte_To_Buffer(unused,0);
    Int_To_Buffer(modes,12);      // 1 for unauthenticated, 2 for authenticated and 4 for encrypted
    Byte_To_Buffer(challenge,16); // random sequence of octets generated by the server
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendCSSG(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
    System.out.println("\nConnectionSetupServerGreeting to "+socket.toString());
    System.out.println();
  }

}

class ConnectionSetupServerGreetingReceiver extends ConnectionSetupServerGreeting{

  public ConnectionSetupServerGreetingReceiver () {
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_ConnectionSetupServerGreeting(byte[] _buffer) {
    System.arraycopy(_buffer,0,unused,0,12);
    modes = (int)(_buffer[15] & 0x07);     //  Just the last 3 bits of modes 32-bit value are used
                                           //  the server should ignore the first 29 bits.
    System.arraycopy(_buffer,16,challenge,0,16);
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveCSSG(Socket socket){
                                                                // readn represents the number of bytes read
    int readn = receivePacket(socket, buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)
    if(readn != 32){  // Packet length should be 32 bytes
      //closeSocket(socket);
      return false;
    }

    this.create_Datagram_ConnectionSetupServerGreeting(buffer);

    return true;
  }

}

/******************************************************************************/

// This class is used to implement the Connection Setup Client Response

class ConnectionSetupClientResponse extends Packets{

  static final int PACKET_LENGTH = 68;

  public ConnectionSetupClientResponse(int _mode) {       // for unauthenticated mode
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    mode = _mode;
    for (int i=0;i<username.length;i++) username[i]=0;
    for (int i=0;i<token.length;i++) token[i]=0;
    for (int i=0;i<client_IV.length;i++) client_IV[i]=0;
  }

  public ConnectionSetupClientResponse(int _mode, byte _username[], byte _token[], byte _client_IV[]) { // for authenticated and encrypted modes
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    mode = _mode;
    System.arraycopy(_username,0,username,0,_username.length);
    System.arraycopy(_token,0,token,0,_token.length);
    System.arraycopy(_client_IV,0,client_IV,0,_client_IV.length);
  }

  public ConnectionSetupClientResponse(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  int mode; // mode that the client chooses to use during this OWAMP-Control Session
  byte username[] = new byte[16];   // unused in unauthenticated mode
  byte token[] = new byte [32];     // unused in unauthenticated mode
  byte client_IV[] = new byte [16]; // unused in unauthenticated mode

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_ConnectionSetupClientResponse(){
    System.out.println(" Mode "+mode);
    for (int i=0;i<username.length;i++)
        System.out.println(" username "+i+" "+username[i] );
    for (int i=0;i<token.length;i++)
        System.out.println(" token "+i+" "+token[i]);
    for (int i=0;i<client_IV.length;i++)
        System.out.println(" client_IV "+i+" "+client_IV[i]);
  }

}

class ConnectionSetupClientResponseToBeSend extends ConnectionSetupClientResponse{

  public ConnectionSetupClientResponseToBeSend (int _mode) {       // for unauthenticated mode
    super(_mode);
  }

  public ConnectionSetupClientResponseToBeSend (int _mode, byte _username[], byte _token[], byte _client_IV[]) { // for authenticated and encrypted modes
    super(_mode, _username, _token, _client_IV);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    Int_To_Buffer(mode,0); // mode that the client chooses to use during this OWAMP-Control Session
                           // 1 for unauthenticated, 2 for authenticated and 4 for encrypted
    Byte_To_Buffer(username, 4);   // unused in unauthenticated mode
    Byte_To_Buffer(token, 20);     // unused in unauthenticated mode
    Byte_To_Buffer(client_IV, 52); // unused in unauthenticated mode
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendCSCR(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
  }

}

class ConnectionSetupClientResponseReceiver extends ConnectionSetupClientResponse{

  public ConnectionSetupClientResponseReceiver() {
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_ConnectionSetupClientResponse (byte[] _buffer) {
    mode = getIntFromByteArray(_buffer,0);
    System.arraycopy(_buffer,4,username,0,16);
    System.arraycopy(_buffer,20,token,0,32);
    System.arraycopy(_buffer,52,client_IV,0,16);

  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveCSCR(Socket socket){
                                                                // readn represents the number of bytes read
    int readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

    if(readn != PACKET_LENGTH){  // Packet length should be 68 bytes
      //closeSocket(socket);
      return false;
    }

    this.create_Datagram_ConnectionSetupClientResponse(buffer);
    System.out.println("ConnectionSetupClientResponse from "+socket.toString());
    System.out.println();

    return true;
  }

}

/******************************************************************************/

// This class is used to implement the Connection Setup Server Response

class ConnectionSetupServerResponse extends Packets{

  static final int PACKET_LENGTH = 48;

  public ConnectionSetupServerResponse(short _accept, byte[] _server_IV, Timestamp _uptime) { // for unauthenticated, authenticated and encrypted modes
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    for (int c = 0; c < unused_MBZ.length; c++) unused_MBZ[c]=0; // unused_MBZ MUST be zero
    accept = (byte)_accept;
    System.arraycopy(_server_IV,0,server_IV,0,_server_IV.length); // unused for unauthenticated mode
    uptime = _uptime;
    for (int c = 0; c < integrityZeroPadding.length; c++) integrityZeroPadding[c] = 0;
  }

  public ConnectionSetupServerResponse(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  byte unused_MBZ[] = new byte [15];
  byte accept;                        // 0 if server accepts authentication and 1 if not
  byte server_IV[] = new byte [16];   // unused in unauthenticated mode
  Timestamp uptime = new Timestamp(); // time when the current instantiation of the server started operating
  byte integrityZeroPadding[] = new byte [8];

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_ConnectionSetupServerResponse(){
    for (int i=0;i<unused_MBZ.length;i++)
       System.out.println(" unused_MBZ "+i+" "+unused_MBZ[i]);
    System.out.println(" accept " +accept);
    for (int i=0;i<server_IV.length;i++)
       System.out.println(" server_IV "+i+" "+server_IV[i]);
    System.out.println(" integerPartOfSeconds "+unsignedIntToLong(uptime.integerPartOfSeconds));
    System.out.println(" fractionalPartOfSeconds "+uptime.fractionalPartOfSeconds);
    for (int i=0;i<integrityZeroPadding.length;i++)
       System.out.println(" integrityZeroPadding "+i+" "+integrityZeroPadding[i] );
  }

}

class ConnectionSetupServerResponseToBeSend extends ConnectionSetupServerResponse{

  public ConnectionSetupServerResponseToBeSend (short _accept, byte[] _server_IV, Timestamp _uptime) { // for unauthenticated, authenticated and encrypted modes
    super(_accept, _server_IV, _uptime);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    Byte_To_Buffer(unused_MBZ,0);
    buffer[15] = accept;
    Byte_To_Buffer(server_IV,16);
    Int_To_Buffer(uptime.integerPartOfSeconds,32);    // unsigned integer number of seconds elapsed since
                                                      // 0h on 1 January 1900.
    Int_To_Buffer(uptime.fractionalPartOfSeconds,36); // fractional part of a second that has elapsed
                                                      // since 0h on 1 January 1900.
    Byte_To_Buffer(integrityZeroPadding,40);
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendCSSR(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
    System.out.println("ConnectionSetupServerResponse to "+socket.toString());
    System.out.println();
  }

}

class ConnectionSetupServerResponseReceiver extends ConnectionSetupServerResponse{

  public ConnectionSetupServerResponseReceiver () {
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_ConnectionSetupServerResponse (byte[] _buffer) {
    System.arraycopy(_buffer,0,unused_MBZ,0,15);
    accept = _buffer[15];
    System.arraycopy(_buffer,16,server_IV,0,16);
    uptime.integerPartOfSeconds = getIntFromByteArray(_buffer,32);
    uptime.fractionalPartOfSeconds = getIntFromByteArray(_buffer,36);
    System.arraycopy(_buffer,40,integrityZeroPadding,0,8);
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveCSSR(Socket socket){
                                                                // readn represents the number of bytes read
    int readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

    if(readn != PACKET_LENGTH){                // Packet length should be 48 bytes
      return false;
    }
                                                                // Integrity Zero Padding should be all zero
    if(!testIntegrityZeroPadding(buffer,(short)40,(short)8)){   // testIntegrityZeroPadding(byte[] buffer, short pos, short length)
      System.err.println("Error: Error of IntegrityZeroPadding in operation Connection Setup.");
      return false;
    }

    this.create_Datagram_ConnectionSetupServerResponse(buffer);

    return true;
  }

}

/******************************************************************************/

//--------------------//
//  Request Sessions  //
//--------------------//

// This class is used to implement the Request Session command

class RequestSession extends Packets{

  static final int PACKET_LENGTH = 112;

  public RequestSession(short _mbz_iPVN, short _confSender, short _confReceiver, int _numberOfScheduleSlots, int _numberOfPackets, int _senderPort, int _receiverPort, byte _senderAddress[], byte _receiverAddress[], byte _sid[], int _paddingLenght, Timestamp _startTime, Timestamp _timeout, byte _typeP_Descriptor[]) {
    this.buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    type = (byte)1;
    mbz_iPVN = (byte)_mbz_iPVN;
    confSender = (byte)_confSender;
    confReceiver = (byte)_confReceiver;
    numberOfScheduleSlots = _numberOfScheduleSlots;
    numberOfPackets = _numberOfPackets;
    senderPort = _senderPort;
    receiverPort = _receiverPort;
    System.arraycopy(_senderAddress,0,senderAddress,0,_senderAddress.length);
    System.arraycopy(_receiverAddress,0,receiverAddress,0,_receiverAddress.length);
    System.arraycopy(_sid,0,sid,0,_sid.length);
    paddingLenght = _paddingLenght;
    startTime = _startTime;
    timeout = _timeout;
    System.arraycopy(_typeP_Descriptor,0,typeP_Descriptor,0,_typeP_Descriptor.length);
    for (int c = 0; c < mBZ2.length; c++) mBZ2[c] = 0;
    for (int c = 0; c < integrityZeroPadding.length; c++) integrityZeroPadding[c] = 0;
  }

  public RequestSession(){
    this.buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    type = (byte)1;
    mbz_iPVN = (byte)0;
    confSender = (byte)0;
    confReceiver = (byte)0;
    numberOfScheduleSlots = 0;
    numberOfPackets = 0;
    senderPort = 0;
    receiverPort = 0;
    senderAddress = new byte [16];
    receiverAddress = new byte [16];
    sid = new byte [16];
    paddingLenght = 0;
    startTime = new Timestamp();
    timeout = new Timestamp();
    typeP_Descriptor = new byte [4];
    mBZ2 = new byte [8];
    integrityZeroPadding = new byte [16];
  }

  byte type;                               // type = 1 indicates that this is a Request Session command
  byte mbz_iPVN;                           // the IP version numbers for Sender and Receiver.
  byte confSender;                         // 1 for ask the server to configure the Sender. In this case, the
                                           // corresponding Port value should be disregarded by the server.
  byte confReceiver;                       // 1 for ask the server to configure the Receiver. In this case,
                                           // the corresponding Port value should be disregarded by the server.
  int numberOfScheduleSlots;               // number of slot records that go between the two blocks of
                                           // Integrity Zero Padding.
  int numberOfPackets;                     // number of active measurement packets to be sent during this
                                           // OWAMP-Test Session.
  int senderPort;                          // the UDP port OWAMP-Test packets will be sent from. Meaningful
                                           // only if ConfSender is not set.
                                           // Note: When sending/receiving this packet, senderPort should be sent/received as an unsigned short variable.
  int receiverPort;                        // the UDP port OWAMP-Test packets are requested to be sent to.
                                           // Meaningful only if ConfReceiver is not set.
                                           // Note: When sending/receiving this packet, receiverPort should be sent/received as an unsigned short variable.
  byte senderAddress[] = new byte [16];    // Sender Address.
  byte receiverAddress[] = new byte [16];  // Receiver Address.
  byte sid[] = new byte [16];              // Session Identifier. Meaningful only if ConfReceiver is 0.
  int paddingLenght;                       // number of octets to be appended to normal OWAMP-Test packet.
                                           // The minimum data segment length is therefore 14 octets in
                                           // unauthenticated mode, and 32 octets in authenticated and
                                           // encrypted modes.
  Timestamp startTime = new Timestamp();   // time when the session is to be started (but not before
                                           // Start-Sessions command is issued).
  Timestamp timeout = new Timestamp();     // timeout interval. Packets that are delayed by more then
                                           // timeout are considered lost.
  byte typeP_Descriptor[] = new byte [4];  // descriptor of a subset of Type-P space
  byte mBZ2[] = new byte [8];
  byte integrityZeroPadding[] = new byte [16];

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_RequestSession () {
    System.out.println(" type " +type);
    System.out.println(" mbz_iPVN " +mbz_iPVN);
    System.out.println(" confSender " +confSender);
    System.out.println(" confReceiver " +confReceiver);
    System.out.println(" numberOfScheduleSlots " +numberOfScheduleSlots);
    System.out.println(" numberOfPackets " +numberOfPackets);
    System.out.println(" senderPort " +senderPort);
    System.out.println(" receiverPort " +receiverPort);
    System.out.println(" senderAddress ");
    for (int i=0;i<senderAddress.length;i++)
      System.out.print((int)(senderAddress[i] & 0xFF)+".");
    System.out.println();
    System.out.println(" receiverAddress ");
    for (int i=0;i<receiverAddress.length;i++)
      System.out.print((int)(receiverAddress[i] & 0xFF)+".");
    System.out.println();
    for (int i=0;i<sid.length;i++)
      System.out.println(" sid "+i+" "+ (int)(sid[i] & 0xFF));
    System.out.println(" paddingLenght " +paddingLenght);
    System.out.println(" startTime.integerPartOfSeconds " +unsignedIntToLong(startTime.integerPartOfSeconds));
    System.out.println(" startTime.fractionalPartOfSeconds " +startTime.fractionalPartOfSeconds);
    System.out.println(" timeout.integerPartOfSeconds " +unsignedIntToLong(timeout.integerPartOfSeconds));
    System.out.println(" timeout.fractionalPartOfSeconds " +timeout.fractionalPartOfSeconds);
    for (int i=0;i<typeP_Descriptor.length;i++)
      System.out.println(" typeP_Descriptor "+i+" "+typeP_Descriptor[i]);
    for (int i=0;i<mBZ2.length;i++)
      System.out.println(" mBZ2 "+i+" "+mBZ2[i]);
    for (int i=0;i<integrityZeroPadding.length;i++)
      System.out.println(" integrityZeroPadding "+i+" "+integrityZeroPadding[i]);

  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer() {
    buffer[0] = type;
    buffer[1] = mbz_iPVN;
    buffer[2] = confSender;
    buffer[3] = confReceiver;
    Int_To_Buffer(numberOfScheduleSlots, 4);
    Int_To_Buffer(numberOfPackets, 8);
    Short_To_Buffer((short)senderPort, 12);
    Short_To_Buffer((short)receiverPort, 14);
    Byte_To_Buffer(senderAddress, 16);
    Byte_To_Buffer(receiverAddress, 32);
    Byte_To_Buffer(sid, 48);
    Int_To_Buffer(paddingLenght, 64);
    Int_To_Buffer(startTime.integerPartOfSeconds, 68);
    Int_To_Buffer(startTime.fractionalPartOfSeconds, 72);
    Int_To_Buffer(timeout.integerPartOfSeconds, 76);
    Int_To_Buffer(timeout.fractionalPartOfSeconds, 80);
    Byte_To_Buffer(typeP_Descriptor, 84);
    Byte_To_Buffer(mBZ2, 88);
    Byte_To_Buffer(integrityZeroPadding, 96);
  }


}

class RequestSessionToBeSend extends RequestSession{

  public RequestSessionToBeSend(short _mbz_iPVN, short _confSender, short _confReceiver, int _numberOfScheduleSlots, int _numberOfPackets, int _senderPort, int _receiverPort, byte _senderAddress[], byte _receiverAddress[], byte _sid[], int _paddingLenght, Timestamp _startTime, Timestamp _timeout, byte _typeP_Descriptor[]) {
    super(_mbz_iPVN, _confSender, _confReceiver, _numberOfScheduleSlots, _numberOfPackets, _senderPort, _receiverPort, _senderAddress, _receiverAddress, _sid, _paddingLenght, _startTime, _timeout, _typeP_Descriptor);
  }

  public RequestSessionToBeSend(RequestSession dataRS) {
    super(dataRS.mbz_iPVN, dataRS.confSender, dataRS.confReceiver, dataRS.numberOfScheduleSlots, dataRS.numberOfPackets, dataRS.senderPort, dataRS.receiverPort, dataRS.senderAddress, dataRS.receiverAddress, dataRS.sid, dataRS.paddingLenght, dataRS.startTime, dataRS.timeout, dataRS.typeP_Descriptor);
  }

  public RequestSessionToBeSend(int _numberOfScheduleSlots){   // Used in SessionReceiver in Fetch Session
   numberOfScheduleSlots = _numberOfScheduleSlots;
 }

  public RequestSessionToBeSend() {
    super();
  }

  /**
   * This method is used to send the Request Session packet using the TCP Socket
   * socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendRequestSession(Socket socket){

    // Send the RequestSession packet

    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
  }

  /**
   * This method is used to send the complete Request Session (RS + SSD +IZP) command using the TCP Socket
   * socket.
   * @param arrayOfdataSSD ScheduleSlotDescriptionsToBeSend[]
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendCompleteRequestSession(ScheduleSlotDescriptionsToBeSend[] arrayOfdataSSD, Socket socket){

    // Send the RequestSession packet

    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);

    /**
     * The datagram RequestSession is immediately followed by one or more schedule slot
     * descriptions (the number of schedule slots is specified in the `Number of Schedule Slots'
     * field of the RequestSession datagram).
     */

    // Send the ScheduleSlotDescriptions packets

    for(int i=0;i<this.numberOfScheduleSlots;i++){
      arrayOfdataSSD[i].sendSSD(socket);

    }

    // These are immediately followed by Integrity Zero Padding

    // Send the IntegrityZeroPadding packet

    IntegrityZeroPaddingToBeSend dataIZP = new IntegrityZeroPaddingToBeSend();
    dataIZP.sendIZP(socket);

  }

}

class RequestSessionReceiver extends RequestSession{

  public RequestSessionReceiver(short _mbz_iPVN, short _confSender, short _confReceiver, int _numberOfScheduleSlots, int _numberOfPackets, int _senderPort, int _receiverPort, byte _senderAddress[], byte _receiverAddress[], byte _sid[], int _paddingLenght, Timestamp _startTime, Timestamp _timeout, byte _typeP_Descriptor[]) {
    super(_mbz_iPVN, _confSender, _confReceiver, _numberOfScheduleSlots, _numberOfPackets, _senderPort, _receiverPort, _senderAddress, _receiverAddress, _sid, _paddingLenght, _startTime, _timeout, _typeP_Descriptor);
  }

  public RequestSessionReceiver() {
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_RequestSession (byte[] _buffer) {
    type = _buffer[0];
    mbz_iPVN = _buffer[1];
    confSender = _buffer[2];
    confReceiver = _buffer[3];
    numberOfScheduleSlots = getIntFromByteArray(_buffer,4);
    numberOfPackets = getIntFromByteArray(_buffer,8);
    senderPort = unsignedShortToInt(getShortFromByteArray(_buffer,12));
    receiverPort = unsignedShortToInt(getShortFromByteArray(_buffer,14));
    System.arraycopy(_buffer,16,senderAddress,0,16);
    System.arraycopy(_buffer,32,receiverAddress,0,16);
    System.arraycopy(_buffer,48,sid,0,16);
    paddingLenght = getIntFromByteArray(_buffer,64);
    startTime.integerPartOfSeconds = getIntFromByteArray(_buffer,68);
    startTime.fractionalPartOfSeconds = getIntFromByteArray(_buffer,72);
    timeout.integerPartOfSeconds = getIntFromByteArray(_buffer,76);
    timeout.fractionalPartOfSeconds = getIntFromByteArray(_buffer,80);
    System.arraycopy(_buffer,84,typeP_Descriptor,0,4);
    System.arraycopy(_buffer,88,mBZ2,0,8);
    System.arraycopy(_buffer,96,integrityZeroPadding,0,16);
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveRS(Socket socket){
                                                               // readn represents the number of bytes read
    int readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

    if(readn != PACKET_LENGTH){
      System.err.println("Error: Error receiving the packet RS.");
      return false;
    }

    if((int)(buffer[0] & 0xFF) != 1){
      System.err.println("Error: The packet received is not a Request-Session packet.");
      return false;
    }

    // Create the Request Session packet received

    if(!testIntegrityZeroPadding(buffer,(short)96,(short)16)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: IntegrityZeroPadding should be all zeros.");
      return false;
    }

    this.create_Datagram_RequestSession (buffer);

    return true;

  }

  /**
   * This method is used to receive the complete Request Session command using the TCP Socket socket.
   * @param arrayOfdataSSD ScheduleSlotDescriptionsReceiver[] - array used to record the ScheduleSlot
   * Descriptions packets.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveCompleteRequestSession(ScheduleSlotDescriptionsReceiver[] arrayOfdataSSD, Socket socket){

    // In variable byte[] buffer we have the packet Request Session received

    if(!testIntegrityZeroPadding(buffer,(short)96,(short)16)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: Error receiving packet RequestSession (Integrity Zero Padding should be all zero).");
     // closeSocket(socket);
      return false;
    }

    this.create_Datagram_RequestSession (buffer);

    // Receive the ScheduleSlotDescriptions packets

                                // Initialize the list to receive the ScheduleSlotDescriptions packets.

    for (int i=0;i<this.numberOfScheduleSlots;i++) {    // For each one of the ScheduleSlots Packets
      arrayOfdataSSD[i] = new ScheduleSlotDescriptionsReceiver();
      if(!arrayOfdataSSD[i].receiveSSD(socket)){
        System.err.println("Error: Error receiving packets ScheduleSlotDescriptions in operation Request Session.");
        return false;
      }

    }

    // Receive the IntegrityZeroPadding packet

    IntegrityZeroPaddingReceiver izp = new IntegrityZeroPaddingReceiver();

    if(!izp.receiveIZP(socket)){
      System.err.println("Error: Error receiving packet IntegrityZeroPadding in operation Request session.");
      return false;
    }

    return true;

  }


  /**
   * This method is used to send the complete Request Session (RS + SSD +IZP) command received using the TCP Socket
   * socket.
   * @param arrayOfdataSSD ScheduleSlotDescriptions[]
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendTheCompleteRSreceived(ScheduleSlotDescriptions[]
                                         arrayOfdataSSD, Socket socket) {

    // Send the RequestSession packet

    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);

    /**
     * The datagram RequestSession is immediately followed by one or more schedule slot
     * descriptions (the number of schedule slots is specified in the `Number of Schedule Slots'
     * field of the RequestSession datagram).
     */

    // Send the ScheduleSlotDescriptions packets

    for (int i = 0; i < this.numberOfScheduleSlots; i++){
      arrayOfdataSSD[i].datagrams_To_Buffer();
      arrayOfdataSSD[i].sendTCP(socket,arrayOfdataSSD[i].PACKET_LENGTH);
    }

    // These are immediately followed by Integrity Zero Padding

    // Send the IntegrityZeroPadding packet

    IntegrityZeroPaddingToBeSend dataIZP = new IntegrityZeroPaddingToBeSend();
    dataIZP.sendIZP(socket);

  }


}

/******************************************************************************/

// This class is used to implement the Schedule Slot Descriptions

class ScheduleSlotDescriptions extends Packets{

  static final int PACKET_LENGTH = 16;

  public ScheduleSlotDescriptions (short _slotType, Timestamp _slotParameter) {
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    slotType = (byte)_slotType;
    for (int c = 0; c < mBZ.length; c++) mBZ[c] = 0;
    slotParameter = _slotParameter;
  }

  public ScheduleSlotDescriptions (){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  byte slotType;                               // 0 to denote a exponentially distributed pseudo-random
                                               // quantity, and 1 to denote a fixed quantity.
  byte mBZ[] = new byte [7];
  Timestamp slotParameter = new Timestamp();   // The time interval that the sender
                                               // will wait before send the Test packet.
                                               // If slotType=0 this interval is the mean
                                               // value (or 1/lambda if the distribution
                                               // density function is expressed as lambda*exp(-lambda*x)
                                               // for positive values of x). If slotType=1
                                               // this interval is the delay itself. So,
                                               // the time elapsed between packets is
                                               // computed according to this parameter

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_ScheduleSlotDescriptions () {
     System.out.println(" slotType " +slotType);
     for (int i=0;i<mBZ.length;i++)
       System.out.println(" mBZ "+i+" "+mBZ[i]);
     System.out.println(" slotParameter.integerPartOfSeconds " +unsignedIntToLong(slotParameter.integerPartOfSeconds));
     System.out.println(" slotParameter.fractionalPartOfSeconds " +slotParameter.fractionalPartOfSeconds);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer() {
    buffer[0] = slotType;
    Byte_To_Buffer(mBZ, 1);
    Int_To_Buffer(slotParameter.integerPartOfSeconds, 8);
    Int_To_Buffer(slotParameter.fractionalPartOfSeconds, 12);
  }


}

class ScheduleSlotDescriptionsToBeSend extends ScheduleSlotDescriptions{

  public ScheduleSlotDescriptionsToBeSend (short _slotType, Timestamp _slotParameter) {
    super(_slotType, _slotParameter);
  }

  public ScheduleSlotDescriptionsToBeSend (ScheduleSlotDescriptionsReceiver dataSSD) {
    super(dataSSD.slotType, dataSSD.slotParameter);
  }

  public ScheduleSlotDescriptionsToBeSend (ScheduleSlotDescriptions dataSSD) {
    super(dataSSD.slotType, dataSSD.slotParameter);
  }

  public ScheduleSlotDescriptionsToBeSend(){
    super();
  }



  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendSSD(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
  }

}

class ScheduleSlotDescriptionsReceiver extends ScheduleSlotDescriptions{

  public ScheduleSlotDescriptionsReceiver (short _slotType, Timestamp _slotParameter) {
    super(_slotType, _slotParameter);
  }

  public ScheduleSlotDescriptionsReceiver(){
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_ScheduleSlotDescriptions (byte[] _buffer) {
     slotType = _buffer[0];
     System.arraycopy(_buffer,1,mBZ,0,7);
     slotParameter.integerPartOfSeconds = getIntFromByteArray(_buffer,8);
     slotParameter.fractionalPartOfSeconds = getIntFromByteArray(_buffer,12);
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveSSD(Socket socket){

    // Receive the ScheduleSlotDescriptions packets

    int readn;           // number of bytes read

    readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

    if(readn != PACKET_LENGTH){            // Packet length should be 16 bytes
      return false;
    }

    // Create the ScheduleSlotDescriptions packets received

    this.create_Datagram_ScheduleSlotDescriptions(buffer);

    return true;

  }

}


/******************************************************************************/

// This class is used to implement the Integrity Zero Padding

class IntegrityZeroPadding extends Packets{

  static final int PACKET_LENGTH = 16;

  public IntegrityZeroPadding(int packetLength){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    integrityZeroPadding = new byte[packetLength];
    for (int c = 0; c < integrityZeroPadding.length; c++)
      integrityZeroPadding[c] = 0;
  }

  public IntegrityZeroPadding(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    integrityZeroPadding = new byte[PACKET_LENGTH];
    for (int c = 0; c < integrityZeroPadding.length; c++)
      integrityZeroPadding[c] = 0;
  }

  byte[] integrityZeroPadding;

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_IntegrityZeroPadding () {
  for (int i=0;i<integrityZeroPadding.length;i++)
    System.out.println(" integrityZeroPadding "+i+" "+integrityZeroPadding[i]);
  }

}

class IntegrityZeroPaddingToBeSend extends IntegrityZeroPadding{

  public IntegrityZeroPaddingToBeSend(int packetLength){
    super(packetLength);
  }

  public IntegrityZeroPaddingToBeSend(){
    super();
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    Byte_To_Buffer(integrityZeroPadding,0);
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendIZP(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, integrityZeroPadding.length);
  }

}

class IntegrityZeroPaddingReceiver extends IntegrityZeroPadding{

  public IntegrityZeroPaddingReceiver(int packetLength){
    super(packetLength);
  }

  public IntegrityZeroPaddingReceiver(){
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_IntegrityZeroPadding (byte[] _buffer) {
    System.arraycopy(_buffer,0,integrityZeroPadding,0,integrityZeroPadding.length);
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveIZP(Socket socket){

    // Receive the IntegrityZeroPadding packet
                                                                             // readn represents the
                                                                             // number of bytes read.
    int readn = receivePacket(socket,buffer,0,integrityZeroPadding.length);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)
    if(readn != integrityZeroPadding.length){  // Packet length should be izpLength bytes
      return false;
    }

    // Process Integrity Zero Padding

    if(!testIntegrityZeroPadding(buffer,(short)0,(short)integrityZeroPadding.length)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: IntegrityZeroPadding should be all zeros.");
      return false;
    }

    this.create_Datagram_IntegrityZeroPadding(buffer);

    return true;

  }

}

/*****************************************************************************/

// This class is used to implement the Accept Session command

class AcceptSession extends Packets{

  static final int PACKET_LENGTH = 32;

  public AcceptSession(short _accept, int _port, byte _sid[]) {
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    accept = (byte)_accept;
    unused = 0;
    port = _port;
    System.arraycopy(_sid,0,sid,0,_sid.length);
    for (int c = 0; c < integrityZeroPadding.length; c++) integrityZeroPadding[c] = 0;
  }

  public AcceptSession(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  byte accept;  // 0 if server accepts the request, 1 if not
  byte unused;
  int port;     // if both ConfSender and ConfReceiver were set port is unused,
                // if only ConfSender was set port is the port to expect OWAMP-Test
                // packets from , if only ConfReceiver was set port is the port
                // to send OWAMP-Test packets to.
                // Note: When sending/receiving this packet, Port should be sent/received as an unsigned short variable.
  byte[] sid = new byte [16]; // Session Identifier. Meaningful only if ConfReceiver is 0.
  byte[] integrityZeroPadding = new byte [12];

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_AcceptSession () {
    System.out.println(" accept " +accept);
    System.out.println(" unused " +unused);
    System.out.println(" port " +port);
    for (int i=0;i<sid.length;i++)
      System.out.println(" sid "+i+" "+(int)(sid[i] & 0xFF));
    for (int i=0;i<integrityZeroPadding.length;i++)
      System.out.println(" integrityZeroPadding "+i+" "+integrityZeroPadding[i]);
   }

}

class AcceptSessionToBeSend extends AcceptSession{

  public AcceptSessionToBeSend(short _accept, int _port, byte _sid[]) {
    super(_accept, _port, _sid);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    buffer[0] = accept;
    buffer[1] = unused;
    Short_To_Buffer((short)port,2);
    Byte_To_Buffer(sid,4);
    Byte_To_Buffer(integrityZeroPadding,20);
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendAS(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
  }

}

class AcceptSessionReceiver extends AcceptSession{

   public AcceptSessionReceiver(){
     super();
   }

   /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

   public void create_Datagram_AcceptSession(byte[] _buffer){
    accept =_buffer[0];
    unused =_buffer[1];
    port = unsignedShortToInt(getShortFromByteArray(_buffer,2));
    System.arraycopy(_buffer,4,sid,0,16);
    System.arraycopy(_buffer,20,integrityZeroPadding,0,12);
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveAS(Socket socket){
                                                                // readn represents the number of bytes read
    int readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

    if(readn != PACKET_LENGTH){  // Packet length should be 32 bytes
      return false;
    }

    this.create_Datagram_AcceptSession(buffer);

    return true;

  }

}

/******************************************************************************/

//--------------------------//
//  Starting Test Sessions  //
//--------------------------//

// This class is used to implement the Starting Test Sessions command

class StartingTestSessions extends Packets{

  static final int PACKET_LENGTH = 32;

  public StartingTestSessions () {
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    type = (byte)2; // 2 indicates that this is a Start Test Sessions command for (int c = 0; c < 15; c++) unused[c] = 0;
    for (int c = 0; c < unused.length; c++) unused[c] = 0;
    for (int c = 0; c < integrityZeroPadding.length; c++) integrityZeroPadding[c] = 0;
  }

  byte type;
  byte[] unused = new byte[15];
  byte[] integrityZeroPadding = new byte[16];

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_StartingTestSessions () {
    System.out.println(" type " +type);
    for (int i=0;i<unused.length;i++)
    System.out.println(" unused "+i+" " +unused[i]);
    for (int i=0;i<integrityZeroPadding.length;i++)
      System.out.println(" integrityZeroPadding "+i+" "+integrityZeroPadding[i]);
  }

}

class StartingTestSessionsToBeSend extends StartingTestSessions{

  public StartingTestSessionsToBeSend () {
    super();
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    buffer[0] = type;
    Byte_To_Buffer(unused,1);
    Byte_To_Buffer(integrityZeroPadding,16);
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendSTS(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
  }

}

class StartingTestSessionsReceiver extends StartingTestSessions{

  public StartingTestSessionsReceiver () {
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_StartingTestSessions (byte[] _buffer) {
    type = _buffer[0];
    System.arraycopy(_buffer,1,unused,0,15);
    System.arraycopy(_buffer,16,integrityZeroPadding,0,16);
  }

}

/*****************************************************************************/

// This class is used to implement the Start-Ack message

class Start_Ack extends Packets{

  static final int PACKET_LENGTH = 32;

  public Start_Ack(byte _accept){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    accept = _accept;
    for (int c = 0; c < mbz.length; c++) mbz[c] = 0;
    for (int c = 0; c < integrityZeroPadding.length; c++) integrityZeroPadding[c] = 0;
  }

  public Start_Ack(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  byte accept; // 0 if server accepts the request, 1 if not.
  byte[] mbz = new byte[15];
  byte[] integrityZeroPadding = new byte[16];

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_Start_Ack () {
    System.out.println(" accept " +accept);
    for (int i=0;i<mbz.length;i++)
      System.out.println(" unused "+i+" " +mbz[i]);
    for (int i=0;i<integrityZeroPadding.length;i++)
      System.out.println(" integrityZeroPadding "+i+" "+integrityZeroPadding[i]);
  }

}

class Start_AckToBeSend extends Start_Ack{

  public Start_AckToBeSend (byte _accept) {
    super(_accept);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    buffer[0]=accept;
    Byte_To_Buffer(mbz,1);
    Byte_To_Buffer(integrityZeroPadding,16);
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendSA(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
  }

}

class Start_AckReceiver extends Start_Ack{

  public Start_AckReceiver(){
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_Start_Ack (byte[] _buffer) {
    accept = _buffer[0];
    System.arraycopy(_buffer,1,mbz,0,15);
    System.arraycopy(_buffer,16,integrityZeroPadding,0,16);
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveSA(Socket socket){
                                                               // readn represents the number of bytes read
    int readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

    if(readn != PACKET_LENGTH){  // Packet length should be 32 bytes
      return false;
    }

                 // Test the IntegrityZeroPadding packet

    if(!testIntegrityZeroPadding(buffer,(short)16,(short)16)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: IntegrityZeroPadding should be all zeros.");
      return false;
    }

                 // Create datagram Control_AckReceiver

    this.create_Datagram_Start_Ack(buffer);

    return true;

  }

}

/******************************************************************************/

//-----------------//
//  Stop Sessions  //
//-----------------//

// This class is used to implement the Stop Sessions command.

class StopSessions extends Packets{

  static final int PACKET_LENGTH = 16;

  public StopSessions(short accept, int numberOfSessions){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    type = 3; // 3 indicates that this is a Stop Session command.
    this.accept = (byte)accept;
    for (int c = 0; c < mbz1.length; c++) mbz1[c] = 0;
    this.numberOfSessions = numberOfSessions;
    for (int c = 0; c < mbz2.length; c++) mbz2[c] = 0;
  }

  public StopSessions(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  byte type;
  byte accept; // 1 indicates a failure of some sort, 0 indicates normal (but possibly premature) completion. All other values are reserved.
  byte[] mbz1 = new byte[2];
  int numberOfSessions; // number of session description records that immediately follow the Stop-Sessions message
  byte[] mbz2 = new byte[8];

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_StopSessions () {
    System.out.println(" type " + type);
    System.out.println(" accept " + accept);
    for (int i = 0; i < mbz1.length; i++)
      System.out.println(" mbz1 " + i + " " + mbz1[i]);
    System.out.println(" numberOfSessions " + numberOfSessions);
    for (int i = 0; i < mbz2.length; i++)
      System.out.println(" mbz2 " + i + " " +mbz2[i]);
  }

  /**
   * This method is used to Stop all sessions initiated in this OWAMP session, in case of failure.
   * @param socket Socket - Server's or Client's socket
   * @param arrayOfSocketToS_R Socket[] - array of all Sender's or Receiver's socket
   * @param numRS int - number of sessions operating
   * @param send boolean - variable used to determine if the Client/Server should send a StopSession
   * to Server/Client or not.
   * Note: When a StopSession is received from Client/Server with accept equal to 1, the Server/Client
   * should not send it back.
   */

  public static void stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send){

    // Send the StopSession to Client/Server. Accept value is 1 because, in case of failure,
    // the Client/Server should stop all sessions and stop this OWAMP-Control session.

    StopSessionsToBeSend dataSS = new StopSessionsToBeSend( (short)1, 0);

    if((send)&&(dataSS.isSocketConnected(socket)))
      dataSS.sendSS(socket);

    closeSocket(socket); // Close socket connection to Client/Server.

    for (int i = 0; i < numRS; i++) // Stop all sessions
      if (!arrayOfSocketToS_R[i].isClosed()) {
        dataSS.sendSS(arrayOfSocketToS_R[i]);
        closeSocket(arrayOfSocketToS_R[i]); // Close socket connection to Sender/Receiver
      }
  }

}

class StopSessionsToBeSend extends StopSessions{

  public StopSessionsToBeSend (short accept, int numberOfSessions){
    super(accept, numberOfSessions);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    buffer[0] = type;
    buffer[1] = accept; // 1 indicates a failure of some sort, 0 indicates normal (but possibly premature) completion. All other values are reserved.
    Byte_To_Buffer(mbz1,2);
    Int_To_Buffer(numberOfSessions,4); // number of session packets sent records that immediately follow the Stop-Sessions message
    Byte_To_Buffer(mbz2,8);
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendSS(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
  }

  /**
   * This method is used to send a complete Stop Session request (SS + SDR).
   * @param sessionDescriptionRecordsToBeSend SessionDescriptionRecordToBeSend[]
   * @param socket Socket - socket to use to send the StopSession.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean sendCompleteSS(SessionDescriptionRecordToBeSend[] sessionDescriptionRecordsToBeSend, Socket socket){

    // Test socket connection

    if((!socket.isConnected())||(socket.isClosed())){
      System.err.println("Error: The socket connection is no longer available.");
      return false;
    }

    // Send the stopSessions packet

    this.sendSS(socket);

    // Send the sessionDescriptionRecords packets

    for(int i=0;i<sessionDescriptionRecordsToBeSend.length;i++){
      if(sessionDescriptionRecordsToBeSend[i] != null)
        sessionDescriptionRecordsToBeSend[i].sendSDR(socket);
    }

    // Send IZP
    IntegrityZeroPaddingToBeSend izp = new IntegrityZeroPaddingToBeSend(16);   // IZP of 16 bytes
    izp.sendIZP(socket);

    return true;

  }

  /**
   * This method is used to send a complete Stop Session request (SS + SDR).
   * @param sessionDescriptionRecordToBeSend SessionDescriptionRecordToBeSend[]
   * @param socket Socket - socket to use to send the StopSession.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean sendCompleteSS(SessionDescriptionRecordToBeSend sessionDescriptionRecordToBeSend, Socket socket){

    // Test socket connection

    if((!socket.isConnected())||(socket.isClosed())){
      System.err.println("Error: The socket connection is no longer available.");
      return false;
    }

    // Send the stopSessions packet

    this.sendSS(socket);

    // Send the sessionDescriptionRecords packets

    sessionDescriptionRecordToBeSend.sendSDR(socket);

    // Send IZP
    IntegrityZeroPaddingToBeSend izp = new IntegrityZeroPaddingToBeSend(16);   // IZP of 16 bytes
    izp.sendIZP(socket);

    return true;

  }

}

class StopSessionsReceiver extends StopSessions{

  public StopSessionsReceiver(int num){
    numberOfSessions = num;
  }

  public StopSessionsReceiver(){
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   */

  public void create_Datagram_StopSessions () {
    type = buffer[0];
    accept = buffer[1];
    System.arraycopy(buffer,2,mbz1,0,2);
    numberOfSessions = getIntFromByteArray(buffer,4);
    System.arraycopy(buffer,8,mbz2,0,8);
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   * @param timeOutToWait double - timeout to wait for a Stop Sessions packet
   * @return short - it returns -2 if the connection to socket is reset, -1 if a failure occurs, 0 if a
   * SS packet is not received during timeout and 1 if a SS packet is received during timeout.
   */

  public short receiveSS(Socket socket, double timeOutToWait){

    if(socket.isClosed()){
      System.err.println("Error: Error in operation receiveSS. The socket connections is closed or doesn't exits.\n");
      return (short)-1;
    }

    if(timeOutToWait < 0)
      timeOutToWait = 1;

    int readn = 0;  // number of bytes read
    try{
      int oldSocketTimeout = socket.getSoTimeout();
      socket.setSoTimeout((int) timeOutToWait); // Set the time out to wait for a Stop Session packet
      readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)
      socket.setSoTimeout(oldSocketTimeout);  // Set the time out to wait for other Control Session packets
                                              // Ex: Fetch Session in same socket.

      switch(readn){
        case -2:  // The socket connections was reset.
          return (short)readn;
        case -1:  // A failure occured reading from socket.
          return (short)readn;
        case 0:   // Timeout to wait occurred. None SS was received.
          return (short)readn;
      }

    }catch(IOException e){
      System.err.println(e);
      return (short)-1;
    }

    if(readn == PACKET_LENGTH){               // Packet length should be 32 bytes

      if((int)(buffer[0] & 0xFF) == 3){     // Packet Stop Sessions

        // In variable byte[] buffer we have the packet stopSessions received


        // Create the stopSessions packet received

        this.create_Datagram_StopSessions();

        return (short)1;
      }
    }

    System.err.println("Error: Packet received is not a StopSession packet.");

    return (short)-1;

  }


  /**
   * This method is used to receive a complete Stop Session request from socket.
   * @param socket Socket - socket used to receive the SS packet.
   * @return SessionPacketsSentRecordReceiver[] - if a SS is received the array of Session Packets
   * SentRecord received is returned, if not or if  a failure occurs it returns null.
   */

  public SessionDescriptionRecordReceiver[] receiveCompleteSS(Socket socket){

    if(socket.isClosed()){
      System.err.println("Error: Error in operation receiveCompleteSS. The socket connections is closed or doesn't exits.");
      return null;
    }

    // In variable byte[] buffer we have the packet stopSessions received

    SessionDescriptionRecordReceiver[] arrayOfdataSDR = new SessionDescriptionRecordReceiver[this.numberOfSessions];

    switch(this.accept){
      case 0: // normal (but possibly premature) completion

        // Receive the SessionDescriptionRecords packets

        for (int i=0;i<this.numberOfSessions;i++){ // For each one of the SessionDescriptionRecord Packets
          arrayOfdataSDR[i] = new SessionDescriptionRecordReceiver();
          if(!arrayOfdataSDR[i].receiveSDR(socket)){
            return null;
          }
        }

        // Receive the IZP
        IntegrityZeroPaddingReceiver izp = new IntegrityZeroPaddingReceiver(16);   // IZP of 16 bytes
        if(!izp.receiveIZP(socket)){
          System.err.println("Error: Error receiving packet IntegrityZeroPadding in operation Stop-Sessions.");
          return null;
        }

        break;
      case 1:
        /** Failure of some sort
         * Results of all OWAMP-Test sessions spawned by this OWAMP-control session MAY be considered
         * invalid.
         */

        System.err.println("Error: Failure of some sort.");
        return null;
      default:
        /**
         * Results of all OWAMP-Test sessions spawned by this OWAMP-control session MAY be considered
         * invalid.
         */

        System.err.println("Error: The accept value is not valid.");
        return null;
    }

    return arrayOfdataSDR;

  }



}

/******************************************************************************/

// This class is used to implement the Session Description Records + Skip Range Descriptions

class SessionDescriptionRecord extends Packets{

  static final int PACKET_LENGTH = 24;

  public SessionDescriptionRecord(byte sid[], int nextSeqno, int numberOfSkipRanges, int[][] skipRangeDescriptions){
    // buffer is an array which will have the information of the packet.
    // The buffer length should be 24 + 8*numberOfSkipRanges. Each
    // Skip Range Descriptions has only 8 bytes, so the last (possibly full, possibly incomplete)
    // block (16 octets) of data MUST be padded with zeros, if necessary. This ensures that the next
    // session description record starts on a block boundary. So if numberOfSkipRanges is odd the
    // buffer length should be 24 + 8*(numberOfSkipRanges+1).
    buffer = new byte[PACKET_LENGTH+8*(((numberOfSkipRanges % 2)==0)?numberOfSkipRanges:(numberOfSkipRanges+1))];

    System.arraycopy(sid,0,this.sid,0,this.sid.length);
    this.nextSeqno = nextSeqno;
    this.numberOfSkipRanges = numberOfSkipRanges;
    if(numberOfSkipRanges > 0){
      this.skipRangeDescriptions = new int[numberOfSkipRanges][2];
      for(int i=0;i<numberOfSkipRanges;i++){
        this.skipRangeDescriptions[i][0] = skipRangeDescriptions[i][0];
        this.skipRangeDescriptions[i][1] = skipRangeDescriptions[i][1];
      }
    }else
      skipRangeDescriptions = null;
  }

  public SessionDescriptionRecord(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  byte[] sid = new byte [16];  // Session Identifier
  int nextSeqno;           // Next Seqno indicates the next sequence number that would have been
                           // sent from this send session. For completed sessions, this will equal
                           // NumPackets from the Request-Session.
  int numberOfSkipRanges;  // Number of Skip Ranges indicates the number of holes that actually
                           // occurred in the sending process. This is a range of packets that were
                           // never actually sent by the sending process.
                           // For the Skip Range Descriptions:
  int[][] skipRangeDescriptions;   // skipRangeDescriptions[][0] - firstSeqnoSkipped
                                   // skipRangeDescriptions[][1] - lastSeqnoSkipped

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_SessionDescriptionRecord() {
    for (int i = 0; i < sid.length; i++)
      System.out.println(" sid " + i + " " + (int)(sid[i] & 0xFF));
    System.out.println(" nextSeqno " + nextSeqno);
    System.out.println(" numberOfSkipRanges " + numberOfSkipRanges);
    for(int i=0;i<numberOfSkipRanges;i++){
      System.out.println(" firstSeqnoSkipped " + skipRangeDescriptions[i][0]);
      System.out.println(" lastSeqnoSkipped " + skipRangeDescriptions[i][0]);
    }
  }

}

class SessionDescriptionRecordToBeSend extends SessionDescriptionRecord{

  public SessionDescriptionRecordToBeSend (byte sid[], int nextSeqno, int numberOfSkipRanges, int[][] skipRangeDescriptions) {
    super(sid, nextSeqno, numberOfSkipRanges, skipRangeDescriptions);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    Byte_To_Buffer(sid,0);
    Int_To_Buffer(nextSeqno,16);
    Int_To_Buffer(numberOfSkipRanges,20);
    if(numberOfSkipRanges > 0){
      int i = 0;
      for (i = 0; i < numberOfSkipRanges; i++) {
        Int_To_Buffer(skipRangeDescriptions[i][0], 24 + (i) * 8);
        Int_To_Buffer(skipRangeDescriptions[i][1], 28 + (i) * 8);
      }

      // Each Skip Range Descriptions has only 8 bytes, so the last (possibly full, possibly incomplete)
      // block (16 octets) of data MUST be padded with zeros, if necessary (numberOfSkipRanges is odd).
      // This ensures that the next session description record starts on a block boundary.
      if ((numberOfSkipRanges % 2) == 1) {
        Int_To_Buffer(0, 24 + (i) * 8);
        Int_To_Buffer(0, 28 + (i) * 8);
      }
    }
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendSDR(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.buffer.length);
  }

}

class SessionDescriptionRecordReceiver extends SessionDescriptionRecord{

  public SessionDescriptionRecordReceiver(){
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   */

  public void create_Datagram_SessionDescriptionRecord() {
    System.arraycopy(buffer,0,sid,0,16);
    nextSeqno = getIntFromByteArray(buffer,16);
    numberOfSkipRanges = getIntFromByteArray(buffer,20);
    if(numberOfSkipRanges > 0){
      this.skipRangeDescriptions = new int[numberOfSkipRanges][2];
      for (int i = 0; i < numberOfSkipRanges; i++) {
        skipRangeDescriptions[i][0] = getIntFromByteArray(buffer,24 + i*8);
        skipRangeDescriptions[i][1] = getIntFromByteArray(buffer,28 + i*8);
      }
    }
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveSDR(Socket socket){

    // First we should read the SessionDescriptionRecords (24 bytes). In the variable buffer we can get the
    // number of Skip Ranges that follows this packet. If this number is greather than 0, we must read the
    // Skip Range descriptions. Each one of the Skip Range descriptions has 8 bytes, so we should read
    // numberOfSkipRanges * 8 bytes.

    // Read the SessionDescriptionRecords (24 bytes).
                                                               // readn represents the number of bytes read
    int readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

    if(readn != PACKET_LENGTH){                // Packet length should be 32 bytes
      System.err.println("Error: Error receiving packets SessionDescriptionRecords in operation Stop Sessions.");
      return false;
    }

    int numSkipRanges = getIntFromByteArray(buffer,20);  // number of Skip Ranges

    if(numSkipRanges > 0){ // If number of Skip Ranges is greather than 0
      // We must resize the buffer used to receive this packet, to have space to save the Skip ranges description.
      // To do that we'll need an auxiliar buffer to save the current information in the buffer.
      byte[] bufferAux = new byte[buffer.length];
      System.arraycopy(buffer,0,bufferAux,0,buffer.length);
      // Each Skip Range Descriptions has only 8 bytes, so the last (possibly full, possibly incomplete)
      // block (16 octets) of data MUST be padded with zeros, if necessary (numberOfSkipRanges is odd).
      // This ensures that the next session description record starts on a block boundary.
      buffer = new byte[PACKET_LENGTH+8*(((numSkipRanges % 2)==0)?numSkipRanges:(numSkipRanges+1))];
      System.arraycopy(bufferAux,0,buffer,0,PACKET_LENGTH);

      // Read the Skip Range descriptions.
      readn = receivePacket(socket, buffer, PACKET_LENGTH, 8*(((numSkipRanges % 2)==0)?numSkipRanges:(numSkipRanges+1))); // receivePacket (Socket socket, byte[] _buffer, int off, int len)

      if (readn != 8*(((numSkipRanges % 2)==0)?numSkipRanges:(numSkipRanges+1))) { // Packet length should be numSkipRanges*8 bytes
        System.err.println("Error: Error receiving packets SkipRangeDescriptions in operation Stop Sessions.");
        return false;
      }
    }

    this.create_Datagram_SessionDescriptionRecord();

    return true;
  }

}

/******************************************************************************/

//-----------------//
//  Fetch Session  //
//-----------------//

// This class is used to implement the Fetch Session

class FetchSession extends Packets{

  static final int PACKET_LENGTH = 48;

  public FetchSession(int _beginSeq, int _endSeq, byte _sid[]){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    type = 4; // 4 indicates that this is a Fetch Session command.
    for (int c = 0; c < unused.length; c++) unused[c] = 0;
    beginSeq = _beginSeq;
    endSeq = _endSeq;
    sid = _sid;
    for (int c = 0; c < integrityZeroPadding.length; c++) integrityZeroPadding[c] = 0;
  }

  public FetchSession(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    type = 4; // 4 indicates that this is a Fetch Session command.
  }

  byte type;
  byte[] unused = new byte[7];
  int beginSeq; // the sequence number of the first requested packet
  int endSeq;   // the sequence number of the last requested packet. If Begin Seq is all zeros and End Seq is all
                // ones, complete session is said to be requested.
  byte[] sid = new byte [16]; // Session Identifier
  byte[] integrityZeroPadding = new byte[16];

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_FetchSession () {
    System.out.println(" type " + type);
    for (int i = 0; i < unused.length; i++)
      System.out.println(" unused " + i + " " + unused[i]);
    System.out.println(" beginSeq " + beginSeq);
    System.out.println(" endSeq " + endSeq);
    for (int i = 0; i < sid.length; i++)
      System.out.println(" sid " + i + " " + (int)(sid[i] & 0xFF));
    for (int i = 0; i < integrityZeroPadding.length; i++)
      System.out.println(" integrityZeroPadding " + i + " " + integrityZeroPadding[i]);
  }

}

class FetchSessionToBeSend extends FetchSession{

  public FetchSessionToBeSend (int _beginSeq, int _endSeq, byte _sid[]) {
    super(_beginSeq, _endSeq, _sid);
  }

  public FetchSessionToBeSend () {
    super();
  }

  /**
   * This method is used to alter the values of beginSeq, endSeq and sid for this FetchSessionToBeSend instance.
   * Using this we can use the same FetchSessionToBeSend instance to fetch several sessions.
   * @param beginSeq int
   * @param endSeq int
   * @param sid byte[]
   */

  public void fetchSessionToBeSend (int beginSeq, int endSeq, byte sid[]) {
    for (int c = 0; c < unused.length; c++)
      unused[c] = 0;
    this.beginSeq = beginSeq;
    this.endSeq = endSeq;
    this.sid = sid;
    for (int c = 0; c < integrityZeroPadding.length; c++)
      integrityZeroPadding[c] = 0;
  }


  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    buffer[0] = type;
    Byte_To_Buffer(unused,1);
    Int_To_Buffer(beginSeq,8);
    Int_To_Buffer(endSeq,12);
    Byte_To_Buffer(sid,16);
    Byte_To_Buffer(integrityZeroPadding,32);
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendFS(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket,this.PACKET_LENGTH);    // send the fetch session message to the server
  }

}

class FetchSessionReceiver extends FetchSession{

  public FetchSessionReceiver(){
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_FetchSession (byte[] _buffer) {
    type = _buffer[0];
    System.arraycopy(_buffer,1,unused,0,7);
    beginSeq = getIntFromByteArray(_buffer,8);
    endSeq = getIntFromByteArray(_buffer,12);
    System.arraycopy(_buffer,16,sid,0,16);
    System.arraycopy(_buffer,32,integrityZeroPadding,0,16);
  }

  /**
   * This method is used to create the packet using the information in the buffer with the packet received
   * from a TCP socket.
   */

  public boolean receiveFS(){

    // In variable byte[] buffer we have the packet FetchSession received

    // Test the FetchSession packet

    if(!testIntegrityZeroPadding(buffer,(short)32,(short)16)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: Error in receiving FetchSession packet.");
      return false;
    }

    this.create_Datagram_FetchSession(buffer);

    return true;

  }

}


/*****************************************************************************/

// This class is used to implement the Fetch-Ack Message

class Fetch_Ack extends Packets{

  static final int PACKET_LENGTH = 32;

  public Fetch_Ack(byte accept, byte finished, int nextSeqno, int numSkipRanges, int numRecords){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    this.accept = accept;
    this.finished = finished;
    for (int c = 0; c < mbz.length; c++) mbz[c] = 0;
    this.nextSeqno = nextSeqno;
    this.numSkipRanges = numSkipRanges;
    this.numRecords = numRecords;
    for (int c = 0; c < integrityZeroPadding.length; c++) integrityZeroPadding[c] = 0;
  }

  public Fetch_Ack(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  byte accept;             // 0 if server accepts the request, 1 if not.
  byte finished;           // Finished is non-zero if the OWAMP-Test session has terminated.
  byte[] mbz = new byte[2];
  int nextSeqno;           // Next Seqno indicates the next sequence number that would have been
                           // sent from this send session. For completed sessions, this will equal
                           // NumPackets from the Request-Session.  This information is only
                           // available if the session has terminated. If Finished is zero, then
                           // Next Seqno MUST be set to zero by the server.

  int numSkipRanges;       // Number of Skip Ranges indicates the number of holes that actually
                           // occurred in the sending process.  This information is only available
                           // if the session has terminated. If Finished is zero, then Skip Ranges
                           // MUST be set to zero by the server.

  int numRecords;          // Number of Records is the number of packet records that fall within
                           // the requested range.  This number might be less than the Number of
                           // Packets in the reproduction of the Request-Session command because of
                           // a session that ended prematurely or it might be greater because of
                           // duplicates.

  byte[] integrityZeroPadding = new byte[16];


  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_Fetch_Ack () {
    System.out.println(" accept " +accept);
    System.out.println(" finished " +finished);
    for (int i=0;i<mbz.length;i++)
      System.out.println(" unused "+i+" " +mbz[i]);
    System.out.println(" nextSeqno " +nextSeqno);
    System.out.println(" numSkipRanges " +numSkipRanges);
    System.out.println(" numRecords " +numRecords);
    for (int i=0;i<integrityZeroPadding.length;i++)
      System.out.println(" integrityZeroPadding "+i+" "+integrityZeroPadding[i]);
  }

}

class Fetch_AckToBeSend extends Fetch_Ack{

  public Fetch_AckToBeSend (byte accept, byte finished, int nextSeqno, int numSkipRanges, int numRecords) {
    super(accept, finished, nextSeqno, numSkipRanges, numRecords);
  }

  public Fetch_AckToBeSend (byte accept) {
    super(accept, (byte)0, 0, 0, 0);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    buffer[0] = accept;
    buffer[1] = finished;
    Byte_To_Buffer(mbz,2);
    Int_To_Buffer(nextSeqno,4);
    Int_To_Buffer(numSkipRanges,8);
    Int_To_Buffer(numRecords,12);
    Byte_To_Buffer(integrityZeroPadding,16);
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendFA(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket, this.PACKET_LENGTH);
  }

}

class Fetch_AckReceiver extends Fetch_Ack{

  public Fetch_AckReceiver(){
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   */

  public void create_Datagram_Fetch_Ack (byte[] _buffer) {
    accept = _buffer[0];
    finished = _buffer[1];
    System.arraycopy(_buffer,2,mbz,0,2);
    nextSeqno = getIntFromByteArray(_buffer,4);
    numSkipRanges = getIntFromByteArray(_buffer,8);
    numRecords = getIntFromByteArray(_buffer,12);
    System.arraycopy(_buffer,16,integrityZeroPadding,0,16);
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveFA(Socket socket){
                                                                // readn represents the number of bytes read
    int readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

    if(readn != PACKET_LENGTH){  // Packet length should be 32 bytes
      return false;
    }

                 // Test the IntegrityZeroPadding packet

    if(!testIntegrityZeroPadding(buffer,(short)16,(short)16)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: IntegrityZeroPadding should be all zeros.");
      return false;
    }

                 // Create datagram Fetch_AckReceiver

    this.create_Datagram_Fetch_Ack(buffer);

    return true;

  }

}


/******************************************************************************/

//-----------------//
//  OTSessionData  //
//-----------------//

// This class is used to implement the OWAMP Test Session Data considering just these fields: reproduction of the
// Request Session, Zero or more (as specified) Skip Range descriptions, IZP 16 bytes, Zero or more (as specified) packet records
// and IZP 16 bytes.

class OTSessionData extends Packets{

  static final int PACKET_LENGTH = 128;

  public OTSessionData(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  int numberOfPacketRecords;   // number of packets records that will follow

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_OTSessionData (RequestSession dataRS,ScheduleSlotDescriptions[] arrayOfdataSSD) {
    dataRS.process_RequestSession ();
    for(int i = 0;i<arrayOfdataSSD.length;i++)
      arrayOfdataSSD[i].process_Datagram_ScheduleSlotDescriptions();
    IntegrityZeroPadding izpRS = new IntegrityZeroPadding(16);  // IZP of 16 bytes
    izpRS.process_Datagram_IntegrityZeroPadding();
    System.out.println("numberOfPacketRecords "+numberOfPacketRecords);
    IntegrityZeroPadding izp = new IntegrityZeroPadding(12);    // IZP of 12 bytes
    izp.process_Datagram_IntegrityZeroPadding();
  }

}

class OTSessionDataToBeSend extends OTSessionData{

  public OTSessionDataToBeSend (RequestSession _dataRS, ScheduleSlotDescriptions[] _arrayOfdataSSD,  int _numberOfPacketRecords/*, OwampTestPacketRecord[] _arrayOfPacketRecords*/) {
    dataRS = new RequestSessionToBeSend(_dataRS.mbz_iPVN, _dataRS.confSender, _dataRS.confReceiver, _dataRS.numberOfScheduleSlots, _dataRS.numberOfPackets, _dataRS.senderPort, _dataRS.receiverPort, _dataRS.senderAddress, _dataRS.receiverAddress, _dataRS.sid, _dataRS.paddingLenght, _dataRS.startTime, _dataRS.timeout, _dataRS.typeP_Descriptor);   // calls the constructor of RequestSession Class
    arrayOfdataSSD = new ScheduleSlotDescriptionsToBeSend[dataRS.numberOfScheduleSlots];
    for(int i = 0;i < dataRS.numberOfScheduleSlots;i++)
      arrayOfdataSSD[i] = new ScheduleSlotDescriptionsToBeSend(_arrayOfdataSSD[i].slotType, _arrayOfdataSSD[i].slotParameter);
    numberOfPacketRecords = _numberOfPacketRecords;
    izp1 = new IntegrityZeroPaddingToBeSend(12);   // IZP of 12 bytes
  }

  public OTSessionDataToBeSend(){
    super();
  }

  RequestSessionToBeSend dataRS;
  ScheduleSlotDescriptionsToBeSend[] arrayOfdataSSD;
  IntegrityZeroPaddingToBeSend izp1;   // IZP of 12 bytes

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public boolean sendCompleteOTSD(Socket socket){

    if(socket.isClosed()){
      System.err.println("Error: Socket connection is closed.");
      return false;
    }

    // Send the complete reproduction of the RequestSession command
    dataRS.sendCompleteRequestSession(arrayOfdataSSD,socket);

    // Send the number Of Packet Records
    Int_To_Buffer(numberOfPacketRecords,0);
    sendTCP(socket,4);

    // Send IZP1
    izp1.sendIZP(socket);

    return true;
  }

}

class OTSessionDataReceiver extends OTSessionData {

  public OTSessionDataReceiver (){
    super();
    dataRS = new RequestSessionReceiver();
    izp1 = new IntegrityZeroPaddingReceiver(12);   // IZP of 12 bytes
  }

  RequestSessionReceiver dataRS;
  ScheduleSlotDescriptionsReceiver[] arrayOfdataSSD;
  IntegrityZeroPaddingReceiver izp1;   // IZP of 12 bytes

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receiveCompleteOTSessionData(Socket socket){

    if(socket.isClosed()){
      System.err.println("Error: Socket connection is closed.");
      return false;
    }

    // Receive the OWAMP-Test session data from socket

         // Receive the Reproduction of the RS used to start the session

             // Process the complete Request Session reception ( RS + SSD + IZP)

                                                         // readn represents the number of bytes read
     int readn = receivePacket(socket, dataRS.buffer,0,dataRS.PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

     if(readn != dataRS.PACKET_LENGTH){
       System.err.println("Error: Error receiving the Reproduction of the RS of OWAMP-Test session data from "+socket.toString());
       return false;
     }
                            // Initialize the list to receive the ScheduleSlotDescriptions packets.

     int numberOfScheduleSlots = Packets.getIntFromByteArray(dataRS.buffer,4);  // numberOfScheduleSlots to receive
     arrayOfdataSSD = new ScheduleSlotDescriptionsReceiver[numberOfScheduleSlots];


     if(!dataRS.receiveCompleteRequestSession(arrayOfdataSSD,socket)){
       System.err.println("Error: Error processing complete Request Session in operation OWAMP-Test session.");
       return false;
     }

             // Receive the number of packet records that will follow and IZP

     readn = receivePacket(socket,buffer,0,4);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

     if(readn != 4){
       System.err.println("Error: Error receiving the numberOfPacketRecords of OWAMP-Test session data from "+socket.toString());
       return false;
     }

     numberOfPacketRecords = getIntFromByteArray(buffer,0);

             // Receive the IntegrityZeroPadding packet

     if(!izp1.receiveIZP(socket)){
       System.err.println("Error: Error receiving packet IntegrityZeroPadding in operation Fetch session.");
       return false;
     }

     return true;
  }

}


/******************************************************************************/

// This class is used to implement the Owamp Test Packet Record

class OwampTestPacketRecord extends Packets{

  static final int PACKET_LENGTH = 25;

  public OwampTestPacketRecord(int _sequenceNumber, Timestamp _sendTimestamp, ErrorEstimate _sendError, Timestamp _receiveTimestamp, ErrorEstimate _receiveError, short _ttl) {
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    sequenceNumber = _sequenceNumber;
    sendError = new ErrorEstimate(_sendError.s_z_scale,_sendError.multiplier);
    receiveError = new ErrorEstimate(_receiveError.s_z_scale,_receiveError.multiplier);
    sendTimestamp = new Timestamp(_sendTimestamp.integerPartOfSeconds,_sendTimestamp.fractionalPartOfSeconds);
    receiveTimestamp = new Timestamp(_receiveTimestamp.integerPartOfSeconds,_receiveTimestamp.fractionalPartOfSeconds);
    ttl = (byte) _ttl;
  }

  public OwampTestPacketRecord(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  int sequenceNumber; // sequence number. Sequence number start with 0 and are incremented by 1 for each subsequent packet
  Timestamp sendTimestamp = new Timestamp(); // send timestamp
  Timestamp receiveTimestamp = new Timestamp(); // receive timestamp
  ErrorEstimate sendError = new ErrorEstimate(); // send timestamp error estimate
  ErrorEstimate receiveError = new ErrorEstimate(); // receive timestamp error estimate
  byte ttl;     // time to live

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_OwampTestPacketRecord () {
    System.out.println(" sequenceNumber " + sequenceNumber);
    System.out.println(" sendTimestamp.integerPartOfSeconds " + unsignedIntToLong(sendTimestamp.integerPartOfSeconds));
    System.out.println(" sendTimestamp.fractionalPartOfSeconds " + sendTimestamp.fractionalPartOfSeconds);
    System.out.println(" receiveTimestamp.integerPartOfSeconds " + unsignedIntToLong(receiveTimestamp.integerPartOfSeconds));
    System.out.println(" receiveTimestamp.fractionalPartOfSeconds " + receiveTimestamp.fractionalPartOfSeconds);
    System.out.println(" sendError.s_z_scale " + (int)(sendError.s_z_scale & 0xFF));
    System.out.println(" sendError.multiplier " + sendError.multiplier);
    System.out.println(" receiveError.s_z_scale " + (int)(receiveError.s_z_scale & 0xFF));
    System.out.println(" receiveError.multiplier " + receiveError.multiplier);
    System.out.println(" TTL " + (int)(ttl & 0xFF));
  }

}

class OwampTestPacketRecordToBeSend extends OwampTestPacketRecord{

  public OwampTestPacketRecordToBeSend(int _sequenceNumber, Timestamp _sendTimestamp, ErrorEstimate _sendError, Timestamp _receiveTimestamp, ErrorEstimate _receiveError, short _ttl){
    super(_sequenceNumber, _sendTimestamp, _sendError, _receiveTimestamp, _receiveError, _ttl);
  }

  public OwampTestPacketRecordToBeSend(){
    super();
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){
    Int_To_Buffer(sequenceNumber,0); // sequence number. Sequence number start with 0 and are incremented by 1 for each subsequent packet
    Int_To_Buffer(sendTimestamp.integerPartOfSeconds,4);  // send timestamp
    Int_To_Buffer(sendTimestamp.fractionalPartOfSeconds,8);
    Int_To_Buffer(receiveTimestamp.integerPartOfSeconds,12); // receive timestamp
    Int_To_Buffer(receiveTimestamp.fractionalPartOfSeconds,16);
    buffer[20] = sendError.s_z_scale; // send timestamp error estimate
    buffer[21] = sendError.multiplier;
    buffer[22] = receiveError.s_z_scale; // receive timestamp error estimate
    buffer[23] = receiveError.multiplier;
    buffer[24] = ttl;
  }

  /**
   * This method is used to send the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to send the packet.
   */

  public void sendOTPR(Socket socket){
    this.datagrams_To_Buffer();
    this.sendTCP(socket,this.PACKET_LENGTH);    // send the fetch session message to the server
 }

 /**
  * Method used to copy the information on OwampTestPacketRecordReceiver to this object.
  * @param dataOTPR OwampTestPacketRecordReceiver
  */

 public void owampTestPacketRecordToBeSend(OwampTestPacketRecordReceiver dataOTPR){
    this.sequenceNumber = dataOTPR.sequenceNumber;
    this.sendTimestamp.timestamp(dataOTPR.sendTimestamp.integerPartOfSeconds,dataOTPR.sendTimestamp.fractionalPartOfSeconds);
    this.receiveTimestamp.timestamp(dataOTPR.receiveTimestamp.integerPartOfSeconds,dataOTPR.receiveTimestamp.fractionalPartOfSeconds);
    this.sendError.errorEstimate(dataOTPR.sendError.s_z_scale,dataOTPR.sendError.multiplier);
    this.receiveError.errorEstimate(dataOTPR.receiveError.s_z_scale,dataOTPR.receiveError.multiplier);
    this.ttl = (byte) dataOTPR.ttl;
  }


}

class OwampTestPacketRecordReceiver extends OwampTestPacketRecord{

  public OwampTestPacketRecordReceiver(){
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer, from offset 'offset'.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   * @param offset int
   */

  public void create_Datagram_OwampTestPacketRecord (byte[] _buffer, int offset) {
    sequenceNumber = getIntFromByteArray(_buffer,0+offset);
    sendTimestamp.integerPartOfSeconds = getIntFromByteArray(_buffer,4+offset);
    sendTimestamp.fractionalPartOfSeconds = getIntFromByteArray(_buffer,8+offset);
    receiveTimestamp.integerPartOfSeconds =  getIntFromByteArray(_buffer,12+offset); // getIntFromByteArray(_buffer,14+offset);
    receiveTimestamp.fractionalPartOfSeconds = getIntFromByteArray(_buffer,16+offset);
    sendError.s_z_scale = _buffer[20+offset];
    sendError.multiplier = _buffer[21+offset];
    receiveError.s_z_scale = _buffer[22+offset];
    receiveError.multiplier = _buffer[23+offset];
    ttl = _buffer[24+offset];
  }

  /**
   * This method is used to receive the packet using the TCP Socket socket.
   * @param socket Socket - TCP Socket used to receive the packet.
   */

  public boolean receivePR(Socket socket){

    // Receive the OwampTestPacketRecord packet

    int readn;           // number of bytes read

    readn = receivePacket(socket,buffer,0,PACKET_LENGTH);  // receivePacket (Socket socket, byte[] _buffer, int off, int len)

    if(readn != PACKET_LENGTH){           // Packet length should be PACKET_LENGTH bytes
      System.err.println("Error: Error receiving packet OwampTestPacketRecord.");
      //closeSocket(socket);
      return false;
    }

    // Create the OwampTestPacketRecord packet received

    this.create_Datagram_OwampTestPacketRecord(buffer,0);

    return true;

  }

}

/******************************************************************************/


                            //----------------//
                            //  OWAMP - Test  //
                            //----------------//


//------------------//
//  Stream Packets  //
//------------------//

// This class is used to implement the Stream Packets Unauthenticated

class StreamPackets_Unauthenticated extends Packets{ // for unauthenticated mode

  static final int PACKET_LENGTH = 14;

  public StreamPackets_Unauthenticated(int _sequenceNumber, Timestamp _timeStamp, ErrorEstimate _errorEstimate, byte _packetPadding[]){ // for unauthenticated mode
    buffer = new byte[PACKET_LENGTH+_packetPadding.length]; // buffer is an array which will have the information of the packet.
    sequenceNumber = _sequenceNumber; // Sequence numbers start with 0 and are incremented by 1 for each subsequent packet
    timeStamp = new Timestamp(_timeStamp.integerPartOfSeconds,_timeStamp.fractionalPartOfSeconds);
    errorEstimate = new ErrorEstimate(_errorEstimate.s_z_scale,_errorEstimate.multiplier);
    packetPadding = new byte[_packetPadding.length];
    System.arraycopy(_packetPadding,0,packetPadding,0,_packetPadding.length); // Packet Padding in OWAMP-Test SHOULD be pseudo-random. However, implementations MUST provide a configuration parameter, an option or a different means of making Packet Padding consist of all zeros
  }

  public StreamPackets_Unauthenticated(int paddingLength){
    buffer = new byte[PACKET_LENGTH+paddingLength];  // buffer is an array which will have the information of the packet.
    packetPadding = new byte[paddingLength];
  }

  int sequenceNumber;
  Timestamp timeStamp = new Timestamp();
  ErrorEstimate errorEstimate = new ErrorEstimate();
  byte[] packetPadding;    // Packet Padding in OWAMP-Test SHOULD be pseudo-random. However, implementations MUST
                           // provide a configuration parameter, an option or a different means of making Packet
                           // Padding consist of all zeros.

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_StreamPackets_Unauthenticated () {
    System.out.println(" sequenceNumber " + sequenceNumber);
    System.out.println(" timeStamp.integerPartOfSeconds " + unsignedIntToLong(timeStamp.integerPartOfSeconds));
    System.out.println(" timeStamp.fractionalPartOfSeconds " + timeStamp.fractionalPartOfSeconds);
    System.out.println(" errorEstimate.s_z_scale " + errorEstimate.s_z_scale);
    System.out.println(" errorEstimate.multiplier " + errorEstimate.multiplier);
    for (int i = 0; i < packetPadding.length; i++)
      System.out.println(" packetPadding " + i + " " + packetPadding[i]);
  }

}

class StreamPackets_UnauthenticatedToBeSend extends StreamPackets_Unauthenticated{

  public StreamPackets_UnauthenticatedToBeSend (int _sequenceNumber, Timestamp _timeStamp, ErrorEstimate _errorEstimate, byte _packetPadding[]){ // for unauthenticated mode
    super(_sequenceNumber, _timeStamp, _errorEstimate, _packetPadding);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   * @param hiResTimer HiResTimer -  represents the high resolution timer used to get the system time.
   */

  public void datagrams_To_Buffer(HiResTimer hiResTimer){ // for unauthenticated mode
    Int_To_Buffer(sequenceNumber,0);      // Sequence numbers start with 0 and are incremented by
                                          // 1 for each subsequent packet
    buffer[12] = errorEstimate.s_z_scale;
    buffer[13] = errorEstimate.multiplier;
    Byte_To_Buffer(packetPadding,14);
    timeStamp.timestamp((double)hiResTimer.microTime(),(short)2);  // Time in microseconds since 1Jan1970
    Int_To_Buffer(timeStamp.integerPartOfSeconds,4);
    Int_To_Buffer(timeStamp.fractionalPartOfSeconds,8);
  }

  /**
   * This method is used to send the packet, using the UDP DatagramSocket.
   * @param ia InetAddress
   * @param ds DatagramSocket
   * @param port int
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the system time.
   */

  public void sendSPU(InetAddress ia, DatagramSocket ds, int port, HiResTimer hiResTimer){
    this.datagrams_To_Buffer(hiResTimer);
    this.sendUDP(ia, ds, port);  // boolean sendUDP(InetAddress ia, DatagramSocket ds, int port)
  }

  /**
   * Method used to update the instance variables of this object.
   * @param _sequenceNumber int
   * @param _timeStamp Timestamp
   * @param _errorEstimate ErrorEstimate
   * @param _packetPadding byte[]
   */

  public void streamPackets_UnauthenticatedToBeSend (int _sequenceNumber, Timestamp _timeStamp, ErrorEstimate _errorEstimate, byte _packetPadding[]){ // for unauthenticated mode
    this.sequenceNumber = _sequenceNumber; // Sequence numbers start with 0 and are incremented by 1 for each subsequent packet
    this.timeStamp.timestamp(_timeStamp.integerPartOfSeconds,_timeStamp.fractionalPartOfSeconds);
    this.errorEstimate.errorEstimate(_errorEstimate.s_z_scale,_errorEstimate.multiplier);
    System.arraycopy(_packetPadding,0,this.packetPadding,0,this.packetPadding.length); // Packet Padding in OWAMP-Test SHOULD be pseudo-random. However, implementations MUST provide a configuration parameter, an option or a different means of making Packet Padding consist of all zeros
  }

}

class StreamPackets_UnauthenticatedReceiver extends StreamPackets_Unauthenticated{

  public StreamPackets_UnauthenticatedReceiver(int paddingLength){
    super(paddingLength);
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   */

  public void create_Datagram_StreamPackets_Unauthenticated () {
    sequenceNumber = getIntFromByteArray(this.buffer,0);
    timeStamp.integerPartOfSeconds = getIntFromByteArray(this.buffer,4);
    timeStamp.fractionalPartOfSeconds = getIntFromByteArray(this.buffer,8);
    errorEstimate.s_z_scale = this.buffer[12];
    errorEstimate.multiplier = this.buffer[13];
    System.arraycopy(this.buffer, 14, packetPadding, 0, packetPadding.length);
  }

}


/******************************************************************************/

// This class is used to implement the Stream Packets Authenticated or Encrypted

class StreamPackets_Authenticated_or_Encrypted extends Packets{ // for authenticated or encrypted modes

  static final int PACKET_LENGTH = 32;

  public StreamPackets_Authenticated_or_Encrypted(int _sequenceNumber, Timestamp _timeStamp, ErrorEstimate _errorEstimate, byte _packetPadding[]){ // for unauthenticated mode
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    sequenceNumber = _sequenceNumber; // Sequence numbers start with 0 and are incremented by 1 for each subsequent packet
    timeStamp = new Timestamp(_timeStamp.integerPartOfSeconds,_timeStamp.fractionalPartOfSeconds);
    errorEstimate = new ErrorEstimate(_errorEstimate.s_z_scale,_errorEstimate.multiplier);
    packetPadding = new byte[_packetPadding.length];
    System.arraycopy(_packetPadding,0,packetPadding,0,_packetPadding.length);
    for (int c = 0; c < integrityZeroPadding1.length; c++) integrityZeroPadding1[c] = 0;
    for (int c = 0; c < integrityZeroPadding2.length; c++) integrityZeroPadding2[c] = 0;
  }

  public StreamPackets_Authenticated_or_Encrypted(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  int sequenceNumber;
  Timestamp timeStamp = new Timestamp();
  ErrorEstimate errorEstimate = new ErrorEstimate();
  byte[] packetPadding;  // Packet Padding in OWAMP-Test SHOULD be pseudo-random. However, implementations MUST
                         // provide a configuration parameter, an option or a different means of making Packet
                         // Padding consist of all zeros.
  byte[] integrityZeroPadding1 = new byte[12];
  byte[] integrityZeroPadding2 = new byte[6];

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_Datagram_StreamPackets_Authenticated_or_Encrypted () {
    System.out.println(" sequenceNumber " + sequenceNumber);
    for (int i = 0; i < integrityZeroPadding1.length; i++)
      System.out.println(" integrityZeroPadding1 " + i + " " + integrityZeroPadding1[i]);
    System.out.println(" timeStamp.integerPartOfSeconds " + unsignedIntToLong(timeStamp.integerPartOfSeconds));
    System.out.println(" timeStamp.fractionalPartOfSeconds " + timeStamp.fractionalPartOfSeconds);
    System.out.println(" errorEstimate.s_z_scale " + errorEstimate.s_z_scale);
    System.out.println(" errorEstimate.multiplier " + errorEstimate.multiplier);
    for (int i = 0; i < integrityZeroPadding2.length; i++)
      System.out.println(" integrityZeroPadding2 " + i + " " + integrityZeroPadding2[i]);
    for (int i = 0; i < packetPadding.length; i++)
      System.out.println(" packetPadding " + i + " " + packetPadding[i]);
  }

}

class StreamPackets_Authenticated_or_EncryptedToBeSend extends StreamPackets_Authenticated_or_Encrypted{

  public StreamPackets_Authenticated_or_EncryptedToBeSend (int _sequenceNumber, Timestamp _timeStamp, ErrorEstimate _errorEstimate, byte _packetPadding[]) { // for unauthenticated mode
    super(_sequenceNumber, _timeStamp, _errorEstimate, _packetPadding);
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   */

  public void datagrams_To_Buffer(){  // for authenticated or encrypted modes.
    Int_To_Buffer(sequenceNumber,0);  // Sequence numbers start with 0 and are incremented by
                                      // 1 for each subsequent packet.
    Byte_To_Buffer(integrityZeroPadding1,4);
    Int_To_Buffer(timeStamp.integerPartOfSeconds,16);
    Int_To_Buffer(timeStamp.fractionalPartOfSeconds,20);
    buffer[24] = errorEstimate.s_z_scale;
    buffer[25] = errorEstimate.multiplier;
    Byte_To_Buffer(integrityZeroPadding2,26);
    Byte_To_Buffer(packetPadding,32);
  }

}

class StreamPackets_Authenticated_or_EncryptedReceiver extends StreamPackets_Authenticated_or_Encrypted{

  public StreamPackets_Authenticated_or_EncryptedReceiver () {
    super();
  }

  /**
   * This method is used to create this object using the information stored in variable buffer.
   * @param _buffer byte[] - a byte array with the necessary information to create this object.
   * @param _buffer_length int
   */

  public void create_Datagram_StreamPackets_Authenticated_or_Encrypted (byte[] _buffer, int _buffer_length) {
   sequenceNumber = getIntFromByteArray(_buffer,0);
   System.arraycopy(_buffer,4,integrityZeroPadding1,0,12);
   timeStamp.integerPartOfSeconds = getIntFromByteArray(_buffer,16);
   timeStamp.fractionalPartOfSeconds = getIntFromByteArray(_buffer,20);
   errorEstimate.s_z_scale = _buffer[24];
   errorEstimate.multiplier = _buffer[25];
   System.arraycopy(_buffer,26,integrityZeroPadding2,0,6);
   packetPadding = new byte[_buffer_length-PACKET_LENGTH];
   System.arraycopy(_buffer,32,packetPadding,0,(_buffer_length-PACKET_LENGTH));
 }

}

/******************************************************************************/

// This class is used to implement the Timestamp

class Timestamp extends Packets{

  static final long NTP2UNIX = ((long)(70 * 365 + 17)) * 86400; // Seconds between 1900-01-01 and 1970-01-01
  static final long EXP2POW32 = ((long)0x1 << 32);              // Const value equal to 2 pow 32.


  /**
   * This Constructor receives a time value and according to the epoch value it creates a timestamp object.
   * @param time double
   * @param epoch short - the epoch value: case 0, parameter time is in seconds since 0h 1 Jan 1900; case 1,
   * parameter time is in milliseconds since 0h 1 Jan 1970; case 2, parameter time is in microseconds since
   * 0h 1 Jan 1970.
   */

  public Timestamp(double time, short epoch){

    long secondsSince0h1Jan1900 = 0;
    switch (epoch) {
      case 0: // Parameter time is in seconds since 0h 1 Jan 1900.
        secondsSince0h1Jan1900 = (long) (time * EXP2POW32);
        integerPartOfSeconds = (int) ( (secondsSince0h1Jan1900 >> 32) &
                                      0xFFFFFFFF);
        fractionalPartOfSeconds = (int) ( (secondsSince0h1Jan1900) & 0xFFFFFFFF);
        break;
      case 1: // Parameter time is in milliseconds since 0h 1 Jan 1970.
        secondsSince0h1Jan1900 = (long) (time / 1000) + NTP2UNIX;
        integerPartOfSeconds = longToUnsignedInt(secondsSince0h1Jan1900);
        fractionalPartOfSeconds = longToUnsignedInt(Math.round( ( (Math.round(
            time * 1e6) % 1e9) * EXP2POW32) / 1e9)); // Resolution to nanosecond
        break;
      case 2: // Parameter time is in microseconds since 0h 1 Jan 1970.
        secondsSince0h1Jan1900 = (long) (time / 1000000) + NTP2UNIX;
        integerPartOfSeconds = longToUnsignedInt(secondsSince0h1Jan1900);
        fractionalPartOfSeconds = longToUnsignedInt(Math.round( ( (Math.round(
            time * 1e3) % 1e9) * EXP2POW32) / 1e9)); // Resolution to nanosecond
        break;
      default:
        integerPartOfSeconds = 0;
        fractionalPartOfSeconds = 0;
    }
  }


  /**
   * This method receives the integerPartOfSeconds and the fractionalPartOfSeconds to be recorded in this
   * object.
   * @param _integerPartOfSeconds int
   * @param _fractionalPartOfSeconds int
   */

  public Timestamp(int _integerPartOfSeconds, int _fractionalPartOfSeconds){
    integerPartOfSeconds = _integerPartOfSeconds;
    fractionalPartOfSeconds = _fractionalPartOfSeconds;
  }

  public Timestamp() {

  }


  int integerPartOfSeconds;    // unsigned integer number of seconds elapsed since 0h on 1 January 1900
  int fractionalPartOfSeconds; // fractional part of a second that has elapsed since 0h on 1 January 1900


  /**
   * This method returns a 'double' value representing the time in milliseconds since 0h 1 Jan 1900.
   * @return double - time in milliseconds since 0h 1 Jan 1900.
   */

  public double timeInMillisecondsSince1900(){
    return (((double)((((long)this.integerPartOfSeconds) << 32)+unsignedIntToLong(this.fractionalPartOfSeconds))/EXP2POW32)*1000);
  }


  /**
   * This method returns the time in milliseconds since 0h 1 Jan 1970.
   * @return double - time in milliseconds since 0h 1 Jan 1970.
   */

  public double timeInMillisecondsSince1970(){
    long secondsSince1970 = unsignedIntToLong(this.integerPartOfSeconds) - NTP2UNIX;

    return (secondsSince1970*1000+(Math.round((unsignedIntToLong(this.fractionalPartOfSeconds)*1e9)/EXP2POW32)/1e6));
  }

  /**
   * This method returns the time in microseconds since 0h 1 Jan 1970.
   * @return double - time in microseconds since 0h 1 Jan 1970.
   */

  public double timeInMicrosecondsSince1970(){
    long secondsSince1970 = unsignedIntToLong(this.integerPartOfSeconds) - NTP2UNIX;

    return (secondsSince1970*1000000+(Math.round((unsignedIntToLong(this.fractionalPartOfSeconds)*1e9)/EXP2POW32)/1e3));
  }


  /**
  * This method used to receive a time value and according to the epoch value it creates a timestamp object.
  * @param time double
  * @param epoch short - the epoch value: case 0, parameter time is in seconds since 0h 1 Jan 1900; case 1,
  * parameter time is in milliseconds since 0h 1 Jan 1970; case 2, parameter time is in microseconds since
  * 0h 1 Jan 1970.
  */

  public void timestamp(double time, short epoch){

    long secondsSince0h1Jan1900 = 0;
    switch(epoch){
      case 0: // Parameter time is in seconds since 0h 1 Jan 1900.
        secondsSince0h1Jan1900 = (long)(time*EXP2POW32);
        integerPartOfSeconds = (int)((secondsSince0h1Jan1900 >> 32) & 0xFFFFFFFF);
        fractionalPartOfSeconds = (int)((secondsSince0h1Jan1900 ) & 0xFFFFFFFF);
        break;
      case 1: // Parameter time is in milliseconds since 0h 1 Jan 1970.
        secondsSince0h1Jan1900 = (long)(time / 1000) + NTP2UNIX;
        integerPartOfSeconds = longToUnsignedInt(secondsSince0h1Jan1900);
        fractionalPartOfSeconds = longToUnsignedInt(Math.round(((Math.round(time*1e6) % 1e9) * EXP2POW32) / 1e9)); // Resolution to nanosecond
        break;
      case 2: // Parameter time is in microseconds since 0h 1 Jan 1970.
        secondsSince0h1Jan1900 = (long)(time / 1000000) + NTP2UNIX;
        integerPartOfSeconds = longToUnsignedInt(secondsSince0h1Jan1900);
        fractionalPartOfSeconds = longToUnsignedInt(Math.round(((Math.round(time*1e3) % 1e9) * EXP2POW32) / 1e9)); // Resolution to nanosecond
        break;
      default:
        integerPartOfSeconds = 0;
        fractionalPartOfSeconds = 0;
    }

  }

  /**
   * This method receives the integerPartOfSeconds and the fractionalPartOfSeconds to be recorded in this
   * object.
   * @param _integerPartOfSeconds int
   * @param _fractionalPartOfSeconds int
   */

  public void timestamp(int _integerPartOfSeconds, int _fractionalPartOfSeconds){
    this.integerPartOfSeconds = _integerPartOfSeconds;
    this.fractionalPartOfSeconds = _fractionalPartOfSeconds;
  }

}


/***********************************************************************************************************/

// This class is used to implement the Error Estimate

class ErrorEstimate extends Packets{

  public ErrorEstimate (byte _s_z_scale, byte _multiplier) {
    s_z_scale = _s_z_scale;
    multiplier = _multiplier;
  }

  public ErrorEstimate () {
  }

  byte s_z_scale;
  byte multiplier;

  /* bit s - 1 if the party generating the timestamp has a clock that is synchronized to UTC using an external source (by GPS or NTP), 0 if there is no notion of external synchronization for the time source
   bit z = 0 -> must be set to zero by the sender and ignored by everyone else
   scale - unsigned integer
   multiplier - unsigned integer. The error estimate is equal to multiplier*2^(-32)*2^scale (in seconds). Multiplier must not be set to zero. If multiplier is zero, the packet should be considered corrupt and discarded
   */

  public void errorEstimate(byte _s_z_scale, byte _multiplier) {
    this.s_z_scale = _s_z_scale;
    this.multiplier = _multiplier;
  }

}


/******************************************************************************/


                            //---------------//
                            //  NTP message  //
                            //---------------//


// This class is used to implement the NTP message used on class HiResTimer. The class
// NTP_Message is used to implement a NTP/SNTP client.

class NTP_Message extends Packets{

  static final int PACKET_LENGTH = 48;

  public NTP_Message(){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
  }

  public NTP_Message(byte versionNumber){
    buffer = new byte[PACKET_LENGTH]; // buffer is an array which will have the information of the packet.
    this.versionNumber = versionNumber;
  }


  // Instance variables

  /**
   * Leap Indicator (LI): This is a two-bit code warning of an impending
   * leap second to be inserted/deleted in the last minute of the current
   * day, with bit 0 and bit 1, respectively, coded as follows:
   *
   * LI   Value   Meaning
   * -------------------------------------------------------
   * 00   0       no warning
   * 01   1       last minute has 61 seconds
   * 10   2       last minute has 59 seconds)
   * 11   3       alarm condition (clock not synchronized)
   */
  byte leapIndicator = 0;

  /**
   * Version Number (VN): This is a three-bit integer indicating the
   * NTP/SNTP version number. The version number is 3 for Version 3 (IPv4
   * only) and 4 for Version 4 (IPv4, IPv6 and OSI). If necessary to
   * distinguish between IPv4, IPv6 and OSI, the encapsulating context
   * must be inspected.
   */
  byte versionNumber = 3;

  /**
   * Mode: This is a three-bit integer indicating the mode, with values
   * defined as follows:
   *
   * Mode    Meaning
   * ------------------------------------
   * 0       reserved
   * 1       symmetric active
   * 2       symmetric passive
   * 3       client
   * 4       server
   * 5       broadcast
   * 6       reserved for NTP control message
   * 7       reserved for private use
   *
   * In unicast and anycast modes, the client sets this field to 3
   * (client) in the request and the server sets it to 4 (server) in the
   * reply. In multicast mode, the server sets this field to 5
   * (broadcast).
   */
  byte mode = 3;

  /**
   * Stratum: This is a eight-bit unsigned integer indicating the stratum
   * level of the local clock, with values defined as follows:
   *
   * Stratum   Meaning
   * ----------------------------------------------
   * 0         unspecified or unavailable
   * 1         primary reference (e.g., radio clock)
   * 2-15      secondary reference (via NTP or SNTP)
   * 16-255    reserved
   */
  short stratum = 0;

  /**
   * Poll Interval: This is an eight-bit signed integer indicating the
   * maximum interval between successive messages, in seconds to the
   * nearest power of two. The values that can appear in this field
   * presently range from 4 (16 s) to 14 (16284 s); however, most
   * applications use only the sub-range 6 (64 s) to 10 (1024 s).
   */
  byte pollInterval = 0;

  /**
   * Precision: This is an eight-bit signed integer indicating the
   * precision of the local clock, in seconds to the nearest power of two.
   * The values that normally appear in this field range from -6 for
   * mains-frequency clocks to -20 for microsecond clocks found in some
   * workstations.
   */
  byte precision = 0;

  /**
   * Root Delay: This is a 32-bit signed fixed-point number indicating the
   * total roundtrip delay to the primary reference source, in seconds
   * with fraction point between bits 15 and 16. Note that this variable
   * can take on both positive and negative values, depending on the
   * relative time and frequency offsets. The values that normally appear
   * in this field range from negative values of a few milliseconds to
   * positive values of several hundred milliseconds.
   */
  double rootDelay = 0;

  /**
   * Root Dispersion: This is a 32-bit unsigned fixed-point number
   * indicating the nominal error relative to the primary reference
   * source, in seconds with fraction point between bits 15 and 16. The
   * values that normally appear in this field range from 0 to several
   * hundred milliseconds.
   */
  double rootDispersion = 0;

  /**
   * Reference Identifier: This is a 32-bit bitstring identifying the
   * particular reference source. In the case of NTP Version 3 or Version
   * 4 stratum-0 (unspecified) or stratum-1 (primary) servers, this is a
   * four-character ASCII string, left justified and zero padded to 32
   * bits. In NTP Version 3 secondary servers, this is the 32-bit IPv4
   * address of the reference source. In NTP Version 4 secondary servers,
   * this is the low order 32 bits of the latest transmit timestamp of the
   * reference source. NTP primary (stratum 1) servers should set this
   * field to a code identifying the external reference source according
   * to the following list. If the external reference is one of those
   * listed, the associated code should be used. Codes for sources not
   * listed can be contrived as appropriate.
   *
   * Code     External Reference Source
   * ----------------------------------------------------------------
   * LOCL     uncalibrated local clock used as a primary reference for
   *          a subnet without external means of synchronization
   * PPS      atomic clock or other pulse-per-second source
   *          individually calibrated to national standards
   * ACTS     NIST dialup modem service
   * USNO     USNO modem service
   * PTB      PTB (Germany) modem service
   * TDF      Allouis (France) Radio 164 kHz
   * DCF      Mainflingen (Germany) Radio 77.5 kHz
   * MSF      Rugby (UK) Radio 60 kHz
   * WWV      Ft. Collins (US) Radio 2.5, 5, 10, 15, 20 MHz
   * WWVB     Boulder (US) Radio 60 kHz
   * WWVH     Kaui Hawaii (US) Radio 2.5, 5, 10, 15 MHz
   * CHU      Ottawa (Canada) Radio 3330, 7335, 14670 kHz
   * LORC     LORAN-C radionavigation system
   * OMEG     OMEGA radionavigation system
   * GPS      Global Positioning Service
   * GOES     Geostationary Orbit Environment Satellite
   */
  byte[] referenceIdentifier = {
      0, 0, 0, 0};

  /**
   * Reference Timestamp: This is the time at which the local clock was
   * last set or corrected, in 64-bit timestamp format (in seconds since 0h 1 Jan 1900).
   */
  Timestamp referenceTimestamp = new Timestamp(0,0);

  /**
   * Originate Timestamp: This is the time at which the request departed
   * the client for the server, in 64-bit timestamp format (in seconds since 0h 1 Jan 1900).
   */
  Timestamp originateTimestamp = new Timestamp(0,0);

  /**
   * Receive Timestamp: This is the time at which the request arrived at
   * the server, in 64-bit timestamp format (in seconds since 0h 1 Jan 1900).
   */
  Timestamp receiveTimestamp = new Timestamp(0,0);

  /**
   * Transmit Timestamp: This is the time at which the reply departed the
   * server for the client, in 64-bit timestamp format (in seconds since 0h 1 Jan 1900).
   */
  Timestamp transmitTimestamp = new Timestamp(0,0);


  // Instance methods

  /**
   * This method is used to print the content of the packet on the screen.
   */

  public void process_NTP_Message(){

    System.out.println(" Leap indicator: " + leapIndicator);
    System.out.println(" Version: " + versionNumber);
    System.out.println(" Mode: " + mode);
    System.out.println(" Stratum: " + stratum);
    System.out.println(" Poll: " + pollInterval);
    System.out.println(" Precision: " + precision);
    System.out.println(" Root delay: " + (rootDelay * 1000) + " ms");
    System.out.println(" Root dispersion: " + (rootDispersion * 1000) + " ms");
    System.out.print(" Reference identifier: ");
    /*
     Reference Identifier: This is a 32-bit bitstring identifying the
     particular reference source. In the case of NTP Version 3 or Version
     4 stratum-0 (unspecified) or stratum-1 (primary) servers, this is a
     four-character ASCII string, left justified and zero padded to 32
     bits. In NTP Version 3 secondary servers, this is the 32-bit IPv4
     address of the reference source. In NTP Version 4 secondary servers,
     this is the low order 32 bits of the latest transmit timestamp of the
     reference source.
     */
    if (stratum == 0 || stratum == 1) {
      System.out.println(new String(referenceIdentifier));
    }else if (versionNumber == 3) {
      try{
        System.out.println(InetAddress.getByAddress(referenceIdentifier));
      }catch(Exception e){
        System.err.print(e);
      }
    }else if (versionNumber == 4) {
      Timestamp timestamp = new Timestamp();
      timestamp.integerPartOfSeconds = 0;
      timestamp.fractionalPartOfSeconds = getIntFromByteArray(referenceIdentifier,0);
      System.out.println((timestamp.timeInMillisecondsSince1900()%1000)/1000);  // Prints the value of referenceIdentifier in seconds
    }

    Calendar time = Calendar.getInstance(); // Get actual time. It will be used to represent all Timestamps.

    System.out.print(" Reference timestamp: ");
    time.setTimeInMillis((long)(referenceTimestamp.timeInMillisecondsSince1970()));
    System.out.println(time.getTime()+" (fraction = "+(referenceTimestamp.timeInMillisecondsSince1970()%1000)+" ms)");
    System.out.print(" Originate timestamp: ");
    time.setTimeInMillis((long)(originateTimestamp.timeInMillisecondsSince1970()));
    System.out.println(time.getTime()+" (fraction = "+(originateTimestamp.timeInMillisecondsSince1970()%1000)+" ms)");
    System.out.print(" Receive timestamp: ");
    time.setTimeInMillis((long)(receiveTimestamp.timeInMillisecondsSince1970()));
    System.out.println(time.getTime()+" (fraction = "+(receiveTimestamp.timeInMillisecondsSince1970()%1000)+" ms)");
    System.out.print(" Transmit timestamp: ");
    time.setTimeInMillis((long)(transmitTimestamp.timeInMillisecondsSince1970()));
    System.out.println(time.getTime()+" (fraction = "+(transmitTimestamp.timeInMillisecondsSince1970()%1000)+" ms)");

  }
}

class NTP_MessageToBeSend extends NTP_Message{

  public NTP_MessageToBeSend(byte versionNumber){
    super(versionNumber);
  }

  public NTP_MessageToBeSend(){
    super();
  }

  /**
   * This method is used to store the information of this object in variable buffer before
   * sending it to a socket.
   * @param hiResTimer HiResTimer -  represents the high resolution timer used to get the system time.
   */

  public void datagrams_To_Buffer(HiResTimer hiResTimer){ // for unauthenticated mode

    buffer[0] = (byte) (leapIndicator << 6 | versionNumber << 3 | mode);
    buffer[1] = (byte) stratum;
    buffer[2] = (byte) pollInterval;
    buffer[3] = (byte) precision;

    // root delay is a 32-bit signed fixed-point number
    int rootDelayAux = (int) (rootDelay * 65536.0);
    buffer[4] = (byte) ( (rootDelayAux >> 24) & 0xFF);
    buffer[5] = (byte) ( (rootDelayAux >> 16) & 0xFF);
    buffer[6] = (byte) ( (rootDelayAux >> 8) & 0xFF);
    buffer[7] = (byte) (rootDelayAux & 0xFF);

    // root dispersion is a 32-bit unsigned fixed-point number
    long rootDispersionAux = (long) (rootDispersion * 65536.0);
    buffer[8] = (byte) ( (rootDispersionAux >> 24) & 0xFF);
    buffer[9] = (byte) ( (rootDispersionAux >> 16) & 0xFF);
    buffer[10] = (byte) ( (rootDispersionAux >> 8) & 0xFF);
    buffer[11] = (byte) (rootDispersionAux & 0xFF);

    buffer[12] = referenceIdentifier[0];
    buffer[13] = referenceIdentifier[1];
    buffer[14] = referenceIdentifier[2];
    buffer[15] = referenceIdentifier[3];

    Int_To_Buffer(referenceTimestamp.integerPartOfSeconds,16);
    Int_To_Buffer(referenceTimestamp.fractionalPartOfSeconds,20);
    Int_To_Buffer(originateTimestamp.integerPartOfSeconds,24);
    Int_To_Buffer(originateTimestamp.fractionalPartOfSeconds,28);
    Int_To_Buffer(receiveTimestamp.integerPartOfSeconds,32);
    Int_To_Buffer(receiveTimestamp.fractionalPartOfSeconds,36);
    if(hiResTimer != null)
      transmitTimestamp.timestamp((double)hiResTimer.milliTime(),(short)1);  // Time in milliseconds since 0h 1 Jan 1970
    else  // Set the transmitTimestamp to 0 seconds since 0h 1 Jan 1970
      transmitTimestamp.timestamp((double)0,(short)1);  // Time in milliseconds since 0h 1 Jan 1970
    Int_To_Buffer(transmitTimestamp.integerPartOfSeconds,40);
    Int_To_Buffer(transmitTimestamp.fractionalPartOfSeconds,44);

  }

  /**
   * This method is used to send the packet, using the UDP DatagramSocket.
   * @param ia InetAddress
   * @param ds DatagramSocket
   * @param port int
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the system time.
   * @return boolean - this method returns true if every thing goes ok, otherwise it returns false.
   */

  public boolean sendMessage(InetAddress ia, DatagramSocket ds, int port, HiResTimer hiResTimer){
    this.datagrams_To_Buffer(hiResTimer);
    return this.sendUDP(ia, ds, port);  // boolean sendUDP(InetAddress ia, DatagramSocket ds, int port)
  }

  /**
   * This method is used to send the packet, using the UDP DatagramSocket. In this case
   * the transmitTimestamp is set to 0 seconds since 1Jan1970.
   * @param ia InetAddress
   * @param ds DatagramSocket
   * @param port int
   * @return boolean - this method returns true if every thing goes ok, otherwise it returns false.
   */

  public boolean sendMessage(InetAddress ia, DatagramSocket ds, int port){
    this.datagrams_To_Buffer(null);
    return this.sendUDP(ia, ds, port);  // boolean sendUDP(InetAddress ia, DatagramSocket ds, int port)
  }

}

class NTP_MessageReceiver extends NTP_Message{

  public NTP_MessageReceiver() {
    super();
  }


  /**
   * This method is used to create this object using the information stored in variable buffer.
   * Note: To receive the NTP server response the following lines should be used:
   *
   * NTP_MessageReceiver ntpMessageReceiver = new NTP_MessageReceiver();
   *
   * // Packet used to receive the NTP server response directly to the ntpMessageReceiver's buffer.
   * DatagramPacket incoming =
   *      new DatagramPacket(ntpMessageReceiver.buffer, ntpMessageReceiver.buffer.length, address, 123);
   *
   * // Get response
   * socket.receive(incoming);
   *
   * // Process response
   * ntpMessageReceiver.create_Datagram_NTP_MessageReceived();
   */

  public void create_Datagram_NTP_MessageReceived () {
    leapIndicator = (byte) ( (buffer[0] >> 6) & 0x3);
    versionNumber = (byte) ( (buffer[0] >> 3) & 0x7);
    mode = (byte) (buffer[0] & 0x7);
    stratum = unsignedByteToShort(buffer[1]);
    pollInterval = buffer[2];
    precision = buffer[3];

    // root delay is a 32-bit signed fixed-point number
    rootDelay = (buffer[4] * 256.0) + unsignedByteToShort(buffer[5]) + (unsignedByteToShort(buffer[6]) / 256.0) +
        (unsignedByteToShort(buffer[7]) / 65536.0);

    // root dispersion is a 32-bit unsigned fixed-point number
    rootDispersion = (unsignedByteToShort(buffer[8]) * 256.0) + unsignedByteToShort(buffer[9]) +
        (unsignedByteToShort(buffer[10]) / 256.0) +
        (unsignedByteToShort(buffer[11]) / 65536.0);

    referenceIdentifier[0] = buffer[12];
    referenceIdentifier[1] = buffer[13];
    referenceIdentifier[2] = buffer[14];
    referenceIdentifier[3] = buffer[15];

    referenceTimestamp.integerPartOfSeconds = getIntFromByteArray(this.buffer,16);
    referenceTimestamp.fractionalPartOfSeconds = getIntFromByteArray(this.buffer,20);
    originateTimestamp.integerPartOfSeconds = getIntFromByteArray(this.buffer,24);
    originateTimestamp.fractionalPartOfSeconds = getIntFromByteArray(this.buffer,28);
    receiveTimestamp.integerPartOfSeconds = getIntFromByteArray(this.buffer,32);
    receiveTimestamp.fractionalPartOfSeconds = getIntFromByteArray(this.buffer,36);
    transmitTimestamp.integerPartOfSeconds = getIntFromByteArray(this.buffer,40);
    transmitTimestamp.fractionalPartOfSeconds = getIntFromByteArray(this.buffer,44);

  }

}




