/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the class used to implement the HiResTimer (High resolution timer).
 * This class is used to define an object to read time information with high resolution.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 2.1
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */

import sun.misc.Perf;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.DatagramPacket;
import java.util.Calendar;
import java.io.IOException;

public class HiResTimer extends Thread{

  // Class Variables

  static final int NTPport = 123;         // Port to be used to connect to NTP server.
  static final int TIMEOUT = 120*1000;    // timeout to wait a packet to be received, in ms (2 minutes).


  // Constructors

  /**
   * When a NTP server address is introduced, the HiResTimer will get the time reference directly from the NTP server.
   * In this case the class NTP_Message is used to retrive the time from the NTP server, using SNTP protocol.
   * The current time is computed using the NTP server time reference and the sun.misc.Perf object.
   * @param ntpServerAddress String - NTP server address to be used.
   * @param updateInterval int - The interval duration (in minutes) between two consecutives updates to the time
   * reference.
   */

  public HiResTimer(String ntpServerAddress, int updateInterval) {
    newHiResTimer(ntpServerAddress, updateInterval);
  }


  /**
   * When a NTP server address is introduced, the HiResTimer will get the time reference directly from the NTP server.
   * In this case the class NTP_Message is used to retrive the time from the NTP server, using SNTP protocol.
   * The current time is computed using the NTP server time reference and the sun.misc.Perf object.
   * If the interval duration (in minutes) between two consecutives updates to the time reference is not defined,
   * the default value (2 minutes) will be used.
   * @param ntpServerAddress String - NTP server address to be used.
   */

  public HiResTimer(String ntpServerAddress) {
    newHiResTimer(ntpServerAddress, 2);
  }


  /**
   * When a NTP server address is not introduced, the HiResTimer will get the time reference from the System time
   * (System.currentTimeMillis()). In this case the system should be synchronized (using GPS or NTP).
   * The current time is computed using the time reference and the sun.misc.Perf object.
   * @param updateInterval int - The interval duration (in minutes) between two consecutives updates to the time
   * reference.
   */

  public HiResTimer(int updateInterval){
    newHiResTimer(null, updateInterval);
  }


  /**
   * When a NTP server address is not introduced, the HiResTimer will get the time reference from the System time
   * (System.currentTimeMillis()). In this case the system should be synchronized (using GPS or NTP).
   * The current time is computed using the time reference and the sun.misc.Perf object.
   * If the interval duration (in minutes) between two consecutives updates to the time reference is not defined,
   * the default value (2 minutes) will be used.
   */

  public HiResTimer(){
    newHiResTimer(null, 2);
  }


  // Instance variables

  private InetAddress ntpServerAddress;  // NTP server address to be used.
  private DatagramSocket socketToNTPserver = null;
  private NTP_MessageToBeSend ntpMessageToBeSend = null;
  private NTP_MessageReceiver ntpMessageReceiver = null;
  private DatagramPacket incoming = null;  // Packet used to receive the NTP server response directly to the ntpMessageReceiver's buffer.
  private int updateInterval; // The interval duration (in minutes) between two consecutives updates to time reference.
  private short testOS;       // If 0 the OS is Windows (In this case, the sun.misc.Perf object will be used to define the time).
                              // If 1 the OS is Linux (In this case the JNI (Java Native Interface) to function gettimeofday
                              // should be used to define the time. But, if a NTP server Address is defined it will be used to
                              // get the time reference, instead).
                              // If 2, the sun.misc.Perf object will be used to define the time.

  long[] timeval = new long[2];  /* Used to receive the time from JNI method linuxMicroTime.
                                    The function gettimeofday fills the struct timeval data type with the current system time.
                                    struct timeval{
                                          long int tv_sec;  This represents the number of seconds since the epoch.
                                          long int tv_usec; This is the fractional second value, represented as the number of microseconds.
                                    }
                                    timeval[0] = tv_sec and timeval[1] = tv_usec.
                                  */

  private Perf hiResTimer;    // Represents the sun.misc.Perf object.
  private double freq;
  private double timeReference, counterReference;


  // Instance methods


  /**
   * This method is an auxiliary method used to create a new HiResTimer object.
   * When a NTP server address is introduced, the HiResTimer will get the time reference directly from the NTP server.
   * In this case the class NTP_Message is used to retrive the time from the NTP server, using SNTP protocol.
   * The current time is computed using the NTP server time reference and the sun.misc.Perf object. If ntpServerAddress
   * is null, the system time (System.currentTimeMillis()) is used to get the time reference. In this case the
   * system should be synchronized (using GPS or NTP).
   * @param ntpServerAddress String - NTP server address to be used.
   * @param updateInterval int - The interval duration (in minutes) between two consecutives updates to the time reference.
   */

  public void newHiResTimer(String ntpServerAddress, int updateInterval){

    if(ntpServerAddress != null){
      try {
        this.ntpServerAddress = InetAddress.getByName(ntpServerAddress); // The NTP server address.
      }catch (Exception e) {
        System.err.println("Error: " + e);
        System.exit(1);
      }
    }else
      this.ntpServerAddress = null;

    this.updateInterval = updateInterval;

    hiResTimer = Perf.getPerf();
    freq = hiResTimer.highResFrequency();

    this.testOS = 2;  // 2 - use sun.misc.Perf.

    // Determine the Operating System. If the OS is Windows the the sun.misc.Perf object should be used to define
    // the time, if the OS is Linux and the ntpServerAddress is null the JNI (Java Native Interface) to function
    // gettimeofday should be used to define the time, otherwise the time reference and the sun.misc.Perf object
    // will be used.

    String osName = ((String)System.getProperties().get("os.name")).toLowerCase();

    if (osName.indexOf("windows") > -1) {  // OS is windows.
      this.testOS = 0;  // 0 - OS is windows, use sun.misc.Perf.
    }

    if ((osName.indexOf("linux") > -1)&&(ntpServerAddress == null)) {  // OS is Linux.
      // The JNI (Java Native Interface) to function gettimeofday should be used to define the time.
      this.testOS = 1;  // 1 - OS is Linux.
      try{
        Runtime.getRuntime().load(System.getProperty("user.dir") +"/HiResTimer/JNIgettimeofday/libHiResTimerImp.so");
      }catch (java.lang.UnsatisfiedLinkError e) {
        e.printStackTrace();
        System.err.println("\nError: Error ocurred using JNI (Java Native Interface) to function gettimeofday.");
        System.err.println("The time based on the sun.misc.Perf object will be used to determine System time.");
        System.out.println();
        this.testOS = 2;  // 2 - use sun.misc.Perf.
      }catch(Exception e){
        e.printStackTrace();
        System.err.println("\nError: Error ocurred using JNI (Java Native Interface) to function gettimeofday.");
        System.err.println(
            "The time based on the sun.misc.Perf object will be used to determine System time.");
        System.out.println();
        this.testOS = 2;  // 2 - use sun.misc.Perf.
      }
    }

    if ((this.testOS == 0)||(this.testOS == 2)) { // 0 or 2 - use sun.misc.Perf.

      if(this.ntpServerAddress != null){  // Get time reference directly from a NTP server.
        try{
          socketToNTPserver = new DatagramSocket();
          socketToNTPserver.setSoTimeout(TIMEOUT); // 2 minutes
          ntpMessageToBeSend = new NTP_MessageToBeSend((byte)3);  // Use NTP version 3.
          ntpMessageReceiver = new NTP_MessageReceiver();
          // Packet used to receive the NTP server response directly to the ntpMessageReceiver's buffer.
          incoming = new DatagramPacket(ntpMessageReceiver.buffer, ntpMessageReceiver.buffer.length, this.ntpServerAddress, NTPport);
        }catch (Exception e) {
          System.err.println("Error: " + e);
          socketToNTPserver.close();
          System.exit(1);
        }
      }

      if(!getTimeReference())   // Get the time reference used to determine the system time.
        System.exit(1);

      if(this.ntpServerAddress != null){  // Print time reference
        Calendar time = Calendar.getInstance(); // Get actual time. It will be used to represent reference timestamp.
        System.out.print("Time reference: ");
        time.setTimeInMillis((long)(this.timeReference));
        System.out.println(time.getTime()+" (fraction = "+(this.timeReference%1000)+" ms)\n");
      }

      this.start();  // The time reference will be frequently updated without interfering on the time reading process.
    }

  }


  /**
   * This method is used to run a thread responsible to "frequently" update the time reference without
   * interfering on the time reading process. The update interval is defined by the updateInterval value, in minutes.
   */

  public void run() {

    try {

      Calendar date = Calendar.getInstance();
      long currentTime = 0;
      int minute = 0, nextMinute = 0;

      do {

        // We should guarantee that all OWAMP elements synchronize at same time to get better results, because if
        // Session-Sender and Session-Receiver are synchronized at the same time they will have almoust the same
        // time reference. To do so, we can update the time reference on each "updateInterval" minutes. Considering
        // that one hour is divided by n intervals of duration "updateInterval" minutes, in each interval, the time
        // reference is updated. Having the current time we can determine on which next minute the time reference
        // should be updated using nextMinute = (minute/updateInterval + 1)*updateInterval (use integer division).
        // Having this value we should sleep only for the remaining time.
        currentTime = (long) (this.milliTime());
        date.setTimeInMillis(currentTime);
        minute = date.get(date.MINUTE);
        nextMinute = (minute / updateInterval + 1) * updateInterval;
        date.set(date.get(date.YEAR), date.get(date.MONTH), date.get(date.DATE),
                 date.get(date.HOUR_OF_DAY), nextMinute, 0);
        date.set(date.MILLISECOND, 0);

        sleep(date.getTimeInMillis() - currentTime);

        synchronizeTimeReference();  // If an error ocurres during synchronization process the time reference is not
                                     // not updated. We should not stop the thread. It keeps trying to update the
                                     // time reference on each every "updateInterval".

      }while (true);

    }
    catch (Exception e) {
    }

  }


  /**
   * This method is used to get the time reference used to determine the system time. If the NTP server's address
   * was not introduced the time reference will be get from the system time (System.currentTimeMillis()). In this
   * case the system should be synchronized (using GPS or NTP). If a NTP server's address was introduced the time
   * reference will be acquired directly from the NTP server. In this case the class NTP_Message is used to retrive
   * the time from the NTP server, using SNTP protocol.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean getTimeReference() {

    // During this process the HiResTimer Thread should have maximum priority.
    this.currentThread().setPriority(this.MAX_PRIORITY);

    if (ntpServerAddress == null) {  // Get time reference from the system time (System.currentTimeMillis())
      long timeAux = System.currentTimeMillis();
      do {
        counterReference = hiResTimer.highResCounter();
        timeReference = System.currentTimeMillis();
      }while (timeReference == timeAux);
      // The priority of the HiResTimer Thread should be set to normal again.
      this.currentThread().setPriority(this.NORM_PRIORITY);

      if(counterReference < 0){  // Reset the base values used for both time and counter reference.
        hiResTimer = Perf.getPerf();

        if(hiResTimer.highResCounter() < 0){
          System.err.println("Error: Unable to reset the counter.");
          System.exit(1);
        }

        if (!getTimeReference()) // Get new time and counter reference.
          System.exit(1);
      }

      return true;
    }

    // Acquire the time reference directly from the NTP server

    System.out.println("Getting time reference from NTP server "+ntpServerAddress+" ...");

    try{

      // Send message to NTP server.
      // Note: The timestamps obtained from Windows are limited to a maximum resolution of 10 to 15 milliseconds,
      // depending on the underlying hardware. Therefore, to prevent errors on time reading we send the NTP message
      // with the field transmit timestamp set to zero seconds since 0h 1 Jan 1970. Determining the time interval
      // between send and receive and setting the destinationTimestamp with this value, the localClockOffset will be
      // equal to the NTP server's time.
      if(!ntpMessageToBeSend.sendMessage(ntpServerAddress,socketToNTPserver,NTPport)){ // The transmit timestamp is set to zero seconds since 0h 1 Jan 1970
        System.err.println("Error: Error getting time reference from NTP server "+ntpServerAddress+".");
        // The priority of the HiResTimer Thread should be set to normal again.
        this.currentThread().setPriority(this.NORM_PRIORITY);
        socketToNTPserver.close();
        return false;
      }

      // Immediately record the counter value
      long counterAux = getTicks();

      // Get response
      socketToNTPserver.receive(incoming);

      // Immediately record the incoming timestamp (counter value)
      counterReference = getTicks();   // This will be the counter reference corresponding to time reference.
      timeReference = counterReference - counterAux;
      double destinationTimestamp = ( (double) (timeReference)*1000 / freq);

      // Process response
      ntpMessageReceiver.create_Datagram_NTP_MessageReceived();

      // Calculate local clock offset (corrected, according to RFC2030 errata)
      double localClockOffset =
          ( (ntpMessageReceiver.receiveTimestamp.timeInMillisecondsSince1970() - ntpMessageReceiver.originateTimestamp.timeInMillisecondsSince1970()) +
           (ntpMessageReceiver.transmitTimestamp.timeInMillisecondsSince1970() - destinationTimestamp)) / 2;

      // Set time reference (in milliseconds)
      timeReference = (destinationTimestamp + localClockOffset);

      // The priority of the HiResTimer Thread should be set to normal again.
      this.currentThread().setPriority(this.NORM_PRIORITY);

      // When the time reference is obtained from a NTP server it should be updated to a better value.
      if(!synchronizeTimeReference()){
        socketToNTPserver.close();
        return false;
      }

    }catch (Exception e) {
      // The priority of the HiResTimer Thread should be set to normal again.
      this.currentThread().setPriority(this.NORM_PRIORITY);
      System.err.println("Error: " + e);
      socketToNTPserver.close();
      return false;
    }

    System.out.println("Time reference was defined with success.\n");

    return true;

  }


  /**
   * This method is used to get a better time reference or get a new time reference from an older one.
   * If the NTP server's address was not introduced the time reference will be updated using the system time
   * (System.currentTimeMillis()). In this case the system should be synchronized (using GPS or NTP). If a NTP
   * server's address was introduced an average localClockOffset value is calculated from 10 time request to NTP
   * server. This clock offset is added to the current time reference. The corresponding counter reference is also
   * updated.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean synchronizeTimeReference() {

    // During this process the HiResTimer Thread should have maximum priority.
    this.currentThread().setPriority(this.MAX_PRIORITY);

    if (ntpServerAddress == null) {  // Update time reference from the system time (System.currentTimeMillis())
      long counterReferenceAux = 0, timeReferenceAux = 0, timeAux = 0;
      timeAux = System.currentTimeMillis();
      do {
        counterReferenceAux = hiResTimer.highResCounter();
        timeReferenceAux = System.currentTimeMillis();
      }while (timeReferenceAux == timeAux);
      synchronized(getClass()){  // To prevent time readings before the time reference is updated.
        counterReference = counterReferenceAux;
        timeReference = timeReferenceAux;

        if(counterReference < 0){  // Reset the base values used for both time and counter reference.
          hiResTimer = Perf.getPerf();

          if(hiResTimer.highResCounter() < 0){
            System.err.println("Error: Unable to reset the counter.");
            System.exit(1);
          }

          if (!getTimeReference()) // Get new time and counter reference.
            System.exit(1);
        }
      }
      // The priority of the HiResTimer Thread should be set to normal again.
      this.currentThread().setPriority(this.NORM_PRIORITY);

      return true;
    }

    // Update the time reference directly from the NTP server

    try{
      double destinationTimestamp = 0, localClockOffset = 0, avgOffSet = 0;
      long counterAux = 0;
      int count = 0, countOffSet = 0;

      avgOffSet = 0; // Used to compute the average clock offset to server clock.

      do {
        count++;

        // Send message to NTP server.
        // The transmit timestamp is set to the current time in seconds since 0h 1 Jan 1900.
        if(!ntpMessageToBeSend.sendMessage(ntpServerAddress,socketToNTPserver,NTPport,this)){   // sendMessage(InetAddress ia, DatagramSocket ds, int port, HiResTimer hiResTimer)
          // The priority of the HiResTimer Thread should be set to normal again.
          this.currentThread().setPriority(this.NORM_PRIORITY);
          return false;
        }

        try {
          // Get response from NTP server and immediately set the receiving time (destinationTimestamp).
          socketToNTPserver.receive(incoming);
          counterAux = getTicks();
          destinationTimestamp = (timeReference + (double) (counterAux - counterReference)*1000 / freq);

          // Process response
          ntpMessageReceiver.create_Datagram_NTP_MessageReceived();

          // Calculate local clock offset in milliseconds (corrected, according to RFC2030 errata)
          localClockOffset =
              ( (ntpMessageReceiver.receiveTimestamp.timeInMillisecondsSince1970() - ntpMessageReceiver.originateTimestamp.timeInMillisecondsSince1970()) +
               (ntpMessageReceiver.transmitTimestamp.timeInMillisecondsSince1970() - destinationTimestamp)) / 2;

          avgOffSet = avgOffSet + localClockOffset;
          countOffSet++;
        }catch (Exception e) {  // If the response is not received an exception is thrown (it is ignored since
        }                       // only the received responses will be used to compute avgOffSet).

      }while (count < 10);

      // Set the counter reference and corresponding time reference (in milliseconds)
      synchronized(getClass()){  // To prevent time readings before the time reference is updated.
        counterAux = getTicks();
        timeReference = ( (timeReference + (double) (counterAux - counterReference)*1000 / freq) +
                          avgOffSet / ((countOffSet!=0)?countOffSet:1));
        counterReference = counterAux;
      }

    }catch (Exception e) {
      // The priority of the HiResTimer Thread should be set to normal again.
      this.currentThread().setPriority(this.NORM_PRIORITY);
      System.err.println("Error: "+e);
      return false;
    }

    // The priority of the HiResTimer Thread should be set to normal again.
    this.currentThread().setPriority(this.NORM_PRIORITY);

    return true;

  }


  // Declare native method linuxMicroTime used to call the native method that returns the time in microseconds
  // since 0h 1 Jan 1970, in linux. The native method uses the function gettimeofday to get the system time.
  // This method will be used if the JOWAMP is running in a linux operating system.

  public native void linuxMicroTime(long[] time);


  /**
   * Method used to return the current time in milliseconds.
   * @return double - current time in milliseconds
   */

  public double milliTime(){

    switch(testOS){
      case 1: // OS is Linux, the JNI (Java Native Interface) to function gettimeofday should be used to define the time.
        linuxMicroTime(timeval);
        return (timeval[0]*1000 + (double)timeval[1]/1e3);
      case 0: // OS is Windows
      case 2: // the sun.misc.Perf object will be used to define the time.
        long counterAux = getTicks();
        return (timeReference +((double)((counterAux - counterReference) * 1000)/ freq));
      default:
        return 0;
    }
  }


  /**
   * Method used to return the current time in microseconds.
   * @return double - current time in microseconds
   */

  public double microTime(){

    switch(testOS){
      case 1: // OS is Linux, the JNI (Java Native Interface) to function gettimeofday should be used to define the time.
        linuxMicroTime(timeval);
        return (timeval[0]*1000000 + timeval[1]);
      case 0: // OS is Windows
      case 2: // the sun.misc.Perf object will be used to define the time.
        long counterAux = getTicks();
        return (timeReference*1000 +((double)((counterAux - counterReference) * 1000000) / freq));
      default:
        return 0;
    }
  }


  /**
   * Method used to return the current time in nanoseconds.
   * @return double - current time in nanoseconds
   */

  public double nanoTime(){
    switch(testOS){
      case 1: // OS is Linux, the JNI (Java Native Interface) to function gettimeofday should be used to define the time.
        linuxMicroTime(timeval);
        return (timeval[0]*1000000000L + timeval[1]*1000L);
      case 0: // OS is Windows
      case 2: // the sun.misc.Perf object will be used to define the time.
        long counterAux = getTicks();
        return (timeReference * 1000000 + ( (double) ( (counterAux - counterReference) * 1000000000) / freq));
      default:
        return 0;
    }

  }


  /**
   * Method used to return the clock frequency.
   * @return double - the clock frequency.
   */

  public double getFrequency(){
    return freq;
  }


  /**
   * Method used to return the current clock ticks.
   * @return long - the current clock ticks.
   */

  public long getTicks(){
    long counterAux = hiResTimer.highResCounter();
    if(counterAux < 0){   // Reset the base values used for both time and counter reference.
      synchronized(this){ // To guarantee that the time reference is reseted only once, in case two simultaneous call
                          // to getTicks() are made, the synchronized method is used and we should verify if the counter
                          // still has negative value (the first thread reaching this peace of code resets the counter).
        counterAux = hiResTimer.highResCounter();
        if(counterAux < 0){
          hiResTimer = Perf.getPerf();
          if(hiResTimer.highResCounter() < 0){
            System.err.println("Error: Unable to reset the counter.");
            System.exit(1);
          }
          if (!getTimeReference()) // Get new time and counter reference.
            System.exit(1);
          counterAux = hiResTimer.highResCounter();
        }
      }
    }

    return counterAux;
  }


  /**
   * Method used to get the counter resolution of this timer.
   * @param loops int - number of loops to be used to get the counter resolution.
   * @return double - counter resolution.
   */

  public long getCounterResolution(int loops){

    long before = 0L;
    long after = 0L;
    long smallest = Long.MAX_VALUE;

    for (int i=0; i<loops; ++i){
      before = hiResTimer.highResCounter();
      after = hiResTimer.highResCounter();
      while (before == after)
        after = hiResTimer.highResCounter();
      smallest = Math.min(smallest, after - before);
    }

    return smallest;
  }


  /**
   * Method used to get the clock resolution of this timer in nanoseconds.
   * @param loops int - number of loops to be used to get the clock resolution.
   * @param clockTicks boolean - to use clock ticks (true) or HiResTimer.nanoTime() (false) to get the clock resolution.
   * @return double - clock resolution in nanoseconds.
   */

  public double getClockResolution(int loops, boolean clockTicks){

    double before = 0L;
    double after = 0L;
    double smallest = Long.MAX_VALUE;

    if (clockTicks) {
      for (int i = 0; i < loops; ++i) {
        before = hiResTimer.highResCounter();
        after = hiResTimer.highResCounter();
        while (before == after)
          after = hiResTimer.highResCounter();
        smallest = Math.min(smallest, after - before);
      }

      return (smallest * 1000000000L / freq);

    }else{
      for (int i = 0; i < loops; ++i) {
        before = this.nanoTime();
        after = this.nanoTime();
        while (before == after)
          after = this.nanoTime();
        smallest = Math.min(smallest, after - before);
      }

      return (smallest);
    }

  }


  // Main function

  public static void main(String[] args) throws IOException{

    String ntpServerAddress = null; // The address of the NTP server to be used to get the time reference.
                                    // If not defined the system time will be used to get the time reference.
    int updateInterval = -1; // The interval duration (in minutes) between two consecutives updates to time reference.

    if ( (args.length == 1) &&
        ( (args[0].compareToIgnoreCase("-?") == 0) ||
         (args[0].compareToIgnoreCase("-help") == 0))) { // Print the help message
      System.out.println("\nUsage: HiResTimer [-options]\n");
      System.out.println("where options include:");
      System.out.println("\n   -n | -ntpServer           \tThe address of the NTP server to be used to get the time reference. If not defined");
      System.out.println("                               \tthe system time will be used to get the time reference.");
      System.out.println("\n   -u | -updateInterval      \tThe value (in minutes) to be used for the interval time between two consecutives");
      System.out.println(
          "                               \tupdates to the time reference.");
      System.out.println(
          "\n   -? | -help                \tPrints this help message.");
      System.exit(1);
    }


    if (args.length != 0)
      try {

        int numOfArg = 0; // Used to determine if all arguments are ok.

        // option -ntpServer
        for (int i = 0; i < args.length; ) {
          if ( (args[i].compareToIgnoreCase("-n") == 0) ||
              (args[i].compareToIgnoreCase("-ntpServer") == 0)) {
            ntpServerAddress = args[i + 1];
            try { // Test if the NTP server address introduced is a valid address.
              if (InetAddress.getByName(ntpServerAddress).isLoopbackAddress()) { // The IP addresse for a given host name.
                System.err.println(
                    "Error: The NTP server's address read is not a valid address (" +
                    InetAddress.getByName(ntpServerAddress) + ").");
                System.exit(1);
              }
            }
            catch (Exception e) {
              System.err.println(
                  "Error: The NTP server's address read is not a valid address (" +
                  ntpServerAddress + ").");
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -updateInterval
        for (int i = 0; i < args.length; ) {
          if ( (args[i].compareToIgnoreCase("-u") == 0) ||
              (args[i].compareToIgnoreCase("-updateInterval") == 0)) {
            updateInterval = Integer.parseInt(args[i + 1]);
            if (updateInterval < 0) {
              System.err.println(
                  "Error: The value of updateInterval can not be a negative number (" +
                  updateInterval + ").");
              System.err.println();
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        if ( (numOfArg < (double) args.length / 2) ||
            (ntpServerAddress == null) && (updateInterval == -1))
          throw new IllegalArgumentException();

      }
      catch (Exception e) {
        System.err.println(
            "\nIllegalArgumentException in thread main: Usage: OWAMP_Server [-options]\n");
        System.err.println("where options include:");
        System.err.println("\n   -n | -ntpServer           \tThe address of the NTP server to be used to get the time reference. If not defined");
        System.err.println("                               \tthe system time will be used to get the time reference.");
        System.err.println("\n   -u | -updateInterval      \tThe value (in minutes) to be used for the interval time between two consecutives");
        System.err.println(
            "                               \tupdates to the time reference.");
        System.err.println(
            "\n   -? | -help                \tPrints this help message.");
        System.exit(1);
      }

    HiResTimer hiResTimer = null;
    if (updateInterval == -1) { // Use default updateInterval defined on HiResTimer class.
      hiResTimer = new HiResTimer(ntpServerAddress);
    }
    else
      hiResTimer = new HiResTimer(ntpServerAddress, updateInterval);

    // Print the current time in ms since 0h 1 Jan 1970
    System.out.println((long)hiResTimer.milliTime());

    System.exit(0);

  }

}
