/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the class used to implement the OWAMP Session-Receiver.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 2.1
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */


import java.io.*;
import java.net.*;

public class OWAMP_SessionReceiver extends OWAMP_SessionTerminal{

  // Class Variables

  static final int TIMEOUTTOSTART = 10*60*1000;              // timeout to wait a Starting Test Session packet
                                                             // to be received, in ms.
  static final int TIMEOUTTOWAIT = 1;     // timeout to wait a StopSession packet to be received, in ms.

  static final int TWODAYSTIMEOUT = 2*24*60*60*1000;  // At the end of a test session, the receiver should wait for a Stop
                                                      // Session to be received during 2DAYSTIMEOUT (Two days timeout is
                                                      // a good value to prevent problems if we have two test sessions controled
                                                      // by the same OWAMP Control session and one is much longer than the other). (ms)


  // Constructors

  /**
   * Constructor for class OWAMP_SessionReceiver.
   * @param socket Socket - represents the socket connection to the Server.
   * @param portRange int[] - the UDP port range for test sessions.
   * @param upTime Timestamp - uptime is a timestamp representing the time when the current instantiation of
   * the receiver started operating, in milliseconds since 1Jan1970.
   * @param sleepVal int - Used to sleep for sleepVal ms before sending the next packet record packet in Fetch-Session
   * operation. We do this to prevent overload of the Receiver processor PC.
   * @param keepResults boolean - If true the configuration (RS files in receiverRSdir) and the results (SID files in
   * receiverTestResults directory) of the test session will be keeped, after the processing of a
   * Fetch-Session requesting all the results of that test session. Otherwise the configuration and the results are
   * deleted (Used by default).
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the time.
   */

  OWAMP_SessionReceiver(Socket socket, int[] portRange, Timestamp upTime, int sleepVal, boolean keepResults, HiResTimer hiResTimer){
    super(portRange);
    try{
      this.socketToCS = socket;
      socketToCS.setSoTimeout(TIMEOUT);
    }
    catch(IOException ex){
      System.err.println(ex);
    }

    dataRS = null;

    dataSSr = new StopSessionsReceiver();

    buffer = new byte[DEFAULT_PACKET_LENGTH];
    this.upTime = new Timestamp(upTime.integerPartOfSeconds, upTime.fractionalPartOfSeconds);
    this.sleepVal = sleepVal;
    this.deleteFetchedResults = !keepResults;

    element = "Server";                       // The Receiver communicates directly with the Server.
    this.hiResTimer = hiResTimer;

    try{
      serverSocket = new ServerSocket();    // Used to verify if a socket connection is closed
                                            // using method isSocketConnected.
    }catch(IOException e){
      System.err.println(e);
    }

  }

  OWAMP_SessionReceiver(){
    super(null);
    element = "Server";                       // The Receiver communicates directly with the Server.
    try{
      serverSocket = new ServerSocket();    // Used to verify if a socket connection is closed
                                            // using method isSocketConnected.
    }catch(IOException e){
      System.err.println(e);
    }
  }


  // Instance variables

  private static File receiverDir = new File("receiverDir");     // The directory of all RS and results received
                                                                 // on this receiver.
  private static File receiverRSdir = new File("receiverDir"+File.separator+"receiverRSdir");     // The directory of all RS received from
                                                                                                  // servers in the receiver.

  private static File receiverTestResults = new File("receiverDir"+File.separator+"receiverTestResults");   // The directory of all results received from
                                                                                                            // senders in this Receiver.
  private int sleepVal;         // Used to sleep for sleepVal ms before sending the next packet record packet in
                                // Fetch-Session operation. We do this to prevent overload of the Receiver processor PC.
  private boolean deleteFetchedResults = false;   // Used to verify if the results of a session should be deleted after
                                                  // a Fetch-Session of all results is completed.
  private SessionDescriptionRecordReceiver[] arrayOfdataSDRr;  // Used to receive the SDR from Server


  // Instance methods


  /**
   * This method start a new process to receive the test packets.
   */

  public void run(){

    try{
                                     // Process Connection Setup Server Greeting

      if(!connectionSetupResponse(mode, socketToCS, upTime)){    //connectionSetupResponse(int modes, Socket socket, Timestamp upTime)

        System.err.println("Error: Error in processing Connection Setup with Server in socket " +socketToCS);

        closeSocket(socketToCS);

        return;
      }

      int readn = 0;
      while(true){ // The Receiver keeps listening for the incoming packets
        if(socketToCS.isClosed()){
          System.out.println("End of the current Session.");
          System.out.println();
          if((ds != null) && (!ds.isClosed()))
            ds.close();
          return;
        }

        socketToCS.setSoTimeout(0);  // Configure time to wait for packets from Server to infinite

        readn = 0;

        readn = receivePacket(socketToCS, buffer, 0, 1);   // number of bytes read
        socketToCS.setSoTimeout(TIMEOUT);

        if(readn != 1){
          if(isSocketConnected(socketToCS))  // An error occurred reading from socket
            System.err.println("Error: Error receiving Packet.");
          else{
            System.err.println("Error: The socket connection to Server"+
                " on "+socketToCS+" is no longer available. All sessions will be finished.\n");
            System.out.println("End of the current Session.\n");
          }
          closeSocket(socketToCS);
          if((ds != null) && (!ds.isClosed()))
            ds.close();
          return;
        }

        if(!processPacketReceived()){
          System.err.println("Error: Error in processing packet received");
          if((ds != null) && (!ds.isClosed()))
            ds.close();
          closeSocket(socketToCS);
          return;
        }
      }

    }
    catch(IOException ex){
      ex.printStackTrace();
    }
    finally{
      try{
        socketToCS.close();
        if((ds != null) && (!ds.isClosed()))
          ds.close();
      }
      catch(IOException ignored){
      }

      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();
    }
  }


  /**
   * Method to receive the packet Request Session send by a client.
   * To each Request-Session message, an OWAMP Session-Receiver MUST respond with an Accept-Session message.
   * @param accept short - specify if the Session-Receiver accepts the request. In this message, zero in the
   * Accept field means that the Session-Receiver is willing to conduct the session.  A value of 1 indicates
   * rejection of the request.  All other values are reserved.
   * @return boolean - This method returns true if the request operation was ok, otherwise it returns false.
   */

  public boolean requestSessionResponse(byte accept){

    AcceptSessionToBeSend dataAS;

    if(dataRS != null){   // if a RS was already received refuse the RS and return
      // We should receive this RS to unfill the socket.
      // In variable byte[] buffer we have the packet Request Session received
      // Process the complete Request Session reception ( RS + SSD + IZP)
      int numberOfScheduleSlots = Packets.getIntFromByteArray(buffer,4);  // numberOfScheduleSlots to receive
      ScheduleSlotDescriptionsReceiver[] arrayOfdataSSDAux = new ScheduleSlotDescriptionsReceiver[numberOfScheduleSlots];
      new RequestSessionReceiver().receiveCompleteRequestSession(arrayOfdataSSDAux,socketToCS);
      dataAS = new AcceptSessionToBeSend((short)1,(short)0,new byte[16]);
      dataAS.sendAS(socketToCS);
      System.out.println("Accept Session to Server on "+socketToCS);
      System.err.println("Error: A second RS was received in the same socket from Server.");
      System.out.println();
      return false;
    }

    // In variable byte[] buffer we have the packet Request Session received

    // Process the complete Request Session reception ( RS + SSD + IZP)

    dataRS = new RequestSessionReceiver();
    System.arraycopy(buffer,0,dataRS.buffer,0,dataRS.PACKET_LENGTH);

                               // Initialize the list to receive the ScheduleSlotDescriptions packets.

    int numberOfScheduleSlots = Packets.getIntFromByteArray(buffer,4);  // numberOfScheduleSlots to receive
    arrayOfdataSSD = new ScheduleSlotDescriptionsReceiver[numberOfScheduleSlots];

    if(!dataRS.receiveCompleteRequestSession(arrayOfdataSSD,socketToCS)){
      System.err.println("Error: Error processing complete Request Session, from Server.");
      closeSocket(socketToCS);
      return false;
    }

    //  To each Request-Session message, an OWAMP Receiver MUST respond with an Accept-Session message.

    // Send the AcceptSession packet

    try{

      // Initialize a datagram socket to receive packets. This socket will be only used if this element is a Receiver.
      if ( (dataRS.confReceiver == 1) && (dataRS.confSender == 0))
        if (this.portRange != null) {
          int i = 0;
          for (i = this.portRange[0]; i <= this.portRange[1]; i++) // Find a free port in the port range
            try {
              ds = new DatagramSocket(i); // initialize a datagram socket on a free port of port range to be used.
              break;
            }
            catch (Exception e) {
            }

          if (i > this.portRange[1]) { // All ports in the specified range are binded. We should reject the session.
            System.err.println(
                "Error: The Session-Receiver has no free UDP port to process this test session.");
            System.err.println("Error: All the UDP ports in the range specified when the Session-Receiver was started are binded.");
            System.err.println();
            accept = 1;
            ds = new DatagramSocket(); // initialize a datagram socket on a free port.
          }

        }
        else
          ds = new DatagramSocket(); // initialize a datagram socket on a free port.
      else
        ds = new DatagramSocket(dataRS.receiverPort); // initialize a datagram socket on port 'receiverPort'

    }
    catch(IOException ex){
      System.err.println(ex);
      accept = 1;
      dataAS = new AcceptSessionToBeSend(accept,0,dataRS.sid);
      dataAS.sendAS(socketToCS);
      System.out.println("Accept Session to Server on "+socketToCS);
      System.out.println();
      System.err.println("Error: The Session-Receiver rejected the Request-Session.");
      System.err.println();
      closeSocket(socketToCS);
      return false;

    }

    dataAS = new AcceptSessionToBeSend(accept,ds.getLocalPort(),dataRS.sid);
    dataAS.sendAS(socketToCS);
    System.out.println("Accept Session to Server on "+socketToCS);

    System.out.println();

    if(accept == 1){  // Even in this case we should return true. The session will be terminated because
                      // the socketToCS is closed.
      System.err.println("Error: The Session-Receiver rejected the Request-Session.");
      System.err.println();
      closeSocket(socketToCS);
    }

    return true;

  }


  /**
   * Having requested one or more test sessions and received affirmative Accept-Session responses, an OWAMP
   * client may start the execution of the requested test sessions by sending a Start-Sessions message to
   * the server.
   * @param accept byte - If Accept is 1, the Start-Sessions request was rejected. Zero means that the
   * command was accepted.  All other values are reserved.  The server MAY and the client SHOULD close the
   * connection in the case of a negative response.
   * @return boolean -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean startingTestSessionsResponse(byte accept){ // process Starting Test Sessions Response

    Start_AckToBeSend dataSA;

    if(dataRS == null){ // if a RS wasn't received yet refuse the STS and return.
      System.err.println("Error: A StartingTestSession command was received before a RequestSession command.");
      System.err.println("Error: This session will be terminated.");

                              // Send datagram Start_AckToBeSend.

      dataSA = new Start_AckToBeSend((byte)1);
      dataSA.sendSA(socketToCS);
      System.out.println("Start-Ack to Server on "+socketToCS);
      System.out.println();

      closeSocket(socketToCS);

      return false;
    }

    // In variable byte[] buffer we have the packet startingTestSessions received.
    // Test the startingTestSessions packet.

    StartingTestSessions dataSTSr = new StartingTestSessions();
    if(!dataSTSr.testIntegrityZeroPadding(buffer,(short)16,(short)16)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: Error receiving packet startingTestSessions.");
      return false;
    }

    if(accept == 0){
      dataRStoFile();  // Save the RS to a file.
      switch(mode){
        case 1:
          if(!process_StreamPackets_Unauthenticated())
            System.err.println("Error: Error processing Stream Packets Unauthenticated.");
          break;

      }
    }else{
      // Send datagram Start_AckToBeSend

      dataSA = new Start_AckToBeSend(accept);
      dataSA.sendSA(socketToCS);
      System.out.println("Start-Ack to Server on " + socketToCS);
      System.out.println();

      System.err.println("Error: The StartingTestSession command was rejected.");
      closeSocket(socketToCS);
    }

    closeSocket(socketToCS);   // At this point the session is already finished.

    return true;

  }


  /**
   * Method used by the Session-Receiver to record the dataRS and the arrayOfdataSSD in a file. For each RS we have a different file,
   * which name is the session SID as a string.
   * @return boolean -  This method returns true if everything goes ok, otherwise it returns false.
   */

  public boolean dataRStoFile(){

    try{

      // Test serverRSdir directory, to prevent errors if a user deletes it. If it doesn't exist it should be created.
      if (!receiverRSdir.getParentFile().exists())
        receiverRSdir.getParentFile().mkdir();
      if (!receiverRSdir.exists())
        receiverRSdir.mkdir();

      File outputFileName = new File(receiverRSdir.getPath()+receiverRSdir.separator+owampHexEncode(dataRS.sid)+".rs");  // If outputFile already exists, delete it
      if(outputFileName.exists())
        outputFileName.delete();

      RandomAccessFile outputFile = new RandomAccessFile(outputFileName,"rw");  // Create a new outputFile
      if(!saveRStoFile(dataRS,arrayOfdataSSD,outputFile)){
        outputFile.close();
        return false;
      }

      outputFile.close();

    }
    catch(IOException ex){
      System.err.println(ex);
      return false;
    }

    return true;
  }


  /**
   * This method is used to process the receiving of the Stream Packets Unauthenticated.
   * @return boolean - This method returns true if the request operation was ok, otherwise it returns false.
   */

  public boolean process_StreamPackets_Unauthenticated(){

    int readn = 0; // number of bytes read
    int lowerSequenceNumber = dataRS.numberOfPackets - 1;
    double startTime = dataRS.startTime.timeInMillisecondsSince1970();
    double receivetime = 0;
    byte s_z_scale = (byte)0x3F, multiplier = (byte)0xFF;
    ErrorEstimate receiverError = new ErrorEstimate(s_z_scale,multiplier);
    int packetPaddingLenght = ((dataRS.paddingLenght + StreamPackets_Unauthenticated.PACKET_LENGTH) < _OWP_TESTREC_SIZE)?(_OWP_TESTREC_SIZE - StreamPackets_Unauthenticated.PACKET_LENGTH):dataRS.paddingLenght;

    StreamPackets_UnauthenticatedReceiver dataSPU = new  StreamPackets_UnauthenticatedReceiver(packetPaddingLenght);

    Timestamp receiveT = new Timestamp((double)0, (short)0);

    double timeOut = dataRS.timeout.timeInMillisecondsSince1900();

    /* TimeOut is the amount  of  time  to  wait  for  a  packet to be received before
     * declaring it lost. As such, it is also the amount  of  time  the
     * test session has to stay active after the last packet is sent to
     * be able to count duplicate packets. Because of this the receiver should
     * wait for TimeOut after receiving the last packet to test if it does not
     * receive duplicate packets.
     */

    double totalTimeToWait = totalTimeToWait(dataRS, arrayOfdataSSD);

    // Compute  the send scheduled time of each packet.

    long[] sendScheduledTime = new long[dataRS.numberOfPackets];
    for(int i=0, indexOfSSD = -1;i < dataRS.numberOfPackets;i++, indexOfSSD++){

      if (indexOfSSD > (arrayOfdataSSD.length - 1))
        indexOfSSD = 0;

      if (i != 0) { // i = 0 - first packet
        sendScheduledTime[i] = sendScheduledTime[i-1] + (long) arrayOfdataSSD[indexOfSSD].slotParameter.timeInMillisecondsSince1900();
      }else
        sendScheduledTime[i] = (long) startTime;

    }

    try{
      // Test receiverTestResults directory, to prevent errors if a user deletes it. If it doesn't exist it should be created.
      if (!receiverTestResults.exists())
        receiverTestResults.mkdir();

      String strSid = owampHexEncode(dataRS.sid);

      File outputFileName = new File(receiverTestResults.getPath()+receiverTestResults.separator+strSid+".owp");
      byte accept = 0;
      if(outputFileName.exists()){
        accept = 1;
        System.err.println("Error: The StartingTestSession command was rejected. The receiver results file "+strSid+".owp already exits.");
      }

      // Send datagram Start_AckToBeSend
      Start_AckToBeSend dataSA;
      dataSA = new Start_AckToBeSend(accept);
      dataSA.sendSA(socketToCS);
      System.out.println("Start-Ack to Server on " + socketToCS);
      System.out.println();

      if(accept != 0)
        return false;

      DatagramPacket incoming = new DatagramPacket(dataSPU.buffer,dataSPU.buffer.length);

      long actualTime = (long)hiResTimer.milliTime();

      // Waits for StartTime less one second. The receiver should start waiting for packets before the sender start sending.
      double timeToWait = (startTime > actualTime+1000)?startTime-actualTime-1000:startTime-actualTime;

      // We should add one second to totalTimeToWait because the receiver starts waiting for packets one second before startTime.
      totalTimeToWait = (startTime > actualTime+1000)?totalTimeToWait + 1000:totalTimeToWait;

      switch(waitTimeToWait(timeToWait)){  // Waits for StartTime.
        case 0:  // None SS received
          break;
        case 1:  // SS received
          return true;
        case -1: // Error receiving SS
          return false;
        default:
          return false;
      }

      startTime = (long)hiResTimer.milliTime();

      RandomAccessFile outputFile = new RandomAccessFile(outputFileName,"rw");

      // Save the StartTime and totalTimeToWait in file

      outputFile.writeDouble(startTime);
      outputFile.writeDouble(totalTimeToWait);

      InetSocketAddress senderAddress = new InetSocketAddress(determineSeReAddress(dataRS.mbz_iPVN, dataRS.senderAddress), dataRS.senderPort);
      System.out.println("Test packets from sender on "+senderAddress.toString()+".");
      System.out.println();

      ds.setSoTimeout(((int)timeOut!=0)?(int)timeOut:1);  // to guarantee that the SS is received the ds has to have a timeOut.

      // During reception the receiver Thread should have maximum priority.
      this.currentThread().setPriority(this.MAX_PRIORITY);

      do{  // During totalTimeToWait, receive test packets.

        readn = -1;

        try{
          ds.receive(incoming);                                // read the received packet.
          receivetime = hiResTimer.microTime();  // Record the time of reception.

          // Now we must verify if this packet was received from the Sender and the UDP port negotiated
          // in the Request-Session operation. If not, the packet should be discarded.

          if(incoming.getSocketAddress().equals((SocketAddress)senderAddress)){
            dataSPU.buffer = incoming.getData();                 // receive the incoming packet.
            readn = incoming.getLength();                        // Number of bytes read.
          }
        }catch(IOException ex){
          //System.err.println(ex);
        }

        // Multiplier MUST NOT be set to zero.  If Multiplier is zero,
        // the packet SHOULD be considered corrupt and discarded.

        if ((readn != -1)&&((dataSPU.buffer[13] & 0xFF) != 0)) {

          dataSPU.create_Datagram_StreamPackets_Unauthenticated();
          lowerSequenceNumber = (dataSPU.sequenceNumber < lowerSequenceNumber)?dataSPU.sequenceNumber:lowerSequenceNumber;
          outputFile.writeInt(dataSPU.sequenceNumber);
          outputFile.writeInt(dataSPU.timeStamp.integerPartOfSeconds);
          outputFile.writeInt(dataSPU.timeStamp.fractionalPartOfSeconds);
          outputFile.writeByte(dataSPU.errorEstimate.s_z_scale);
          outputFile.writeByte(dataSPU.errorEstimate.multiplier);

          // Packets not received within the Timeout are considered lost.  They are recorded with their
          // sequence number, presumed send time, and receive time consisting of a string of zero bits.
          // If any of the following is true, the packet MUST be discarded:
          // - Send timestamp is more than Timeout in the past or in the future.
          // - Send timestamp differs by more than Timeout from the time when the packet should have been
          // sent according to its sequence number.

          if (((dataSPU.timeStamp.timeInMillisecondsSince1970() + timeOut) < receivetime/1000)
              || ((dataSPU.timeStamp.timeInMillisecondsSince1970() - timeOut) > receivetime/1000)){
            // - Send timestamp is more than Timeout in the past or in the future.
            receiveT.timestamp((double)0,(short)0);          // Time in seconds since 1Jan1900
          }else{
            if(dataSPU.timeStamp.timeInMillisecondsSince1970() > (sendScheduledTime[dataSPU.sequenceNumber]+timeOut)){
              // - Send timestamp differs by more than Timeout from the time when the packet should have been
              // sent according to its sequence number.
              receiveT.timestamp((double)0,(short)0);          // Time in seconds since 1Jan1900
            }else
              receiveT.timestamp((double)receivetime,(short)2);  // Time in microseconds since 1Jan1970
          }

          outputFile.writeInt(receiveT.integerPartOfSeconds);
          outputFile.writeInt(receiveT.fractionalPartOfSeconds);
          outputFile.writeByte(receiverError.s_z_scale);
          outputFile.writeByte(receiverError.multiplier);
          outputFile.writeByte(255);   // TTL value

        }

        // Test if a Stop Session packet has been received

        switch(testStopSession(dataSPU.timeStamp.timeInMillisecondsSince1970(),lowerSequenceNumber,TIMEOUTTOWAIT)){
          case 0:  // None SS received
            break;
          case 1:  // SS received
            outputFile.close();
            this.currentThread().setPriority(this.NORM_PRIORITY);
            return true;
          case -1: // Error receiving SS
            outputFile.close();
            this.currentThread().setPriority(this.NORM_PRIORITY);
            return false;
          default:
            outputFile.close();
            this.currentThread().setPriority(this.NORM_PRIORITY);
            return false;
        }

        /* Verify the socketToCS (Server). If it is closed before totalTimeToWait, it means that a failure has
         * occurred, possibly a loss of connection to server. The session must be finished. But it can happens
         * that the server sent a SS and closed the connection to Receiver, but the SS had a delay to reach the
         * Receiver. So we should wait for this SS for a timeout before invalidating the results.
         */

        if(!isSocketConnected(socketToCS)){
          outputFile.close();

          // Test if a Stop Session packet has been received (during 2 minutes)

          switch(testStopSession(dataSPU.timeStamp.timeInMillisecondsSince1970(),lowerSequenceNumber,TIMEOUT)){
            case 0:  // None SS received
              break;
            case 1:  // SS received
              this.currentThread().setPriority(this.NORM_PRIORITY);
              return true;
            case -1: // Error receiving SS
              break;
            default:
              break;
          }

          invalidateResults(false);  // invalidateResults(boolean sendSStoServer)
          System.err.println();
          System.err.println("Error: The socket connection to Server is no longer available. This session will be finished.");
          this.currentThread().setPriority(this.NORM_PRIORITY);
          return false;
        }

      }while(((startTime + totalTimeToWait) > ((long)hiResTimer.milliTime()))); // The amount of time to wait
                                                                                // to end the session. Note that
                                                                                // between two packets we can
                                                                                // receive lost or duplicate
                                                                                // packets, so we should receive
                                                                                // any packets during this interval.
      // The priority of the receiver Thread should be set to normal again.
      this.currentThread().setPriority(this.NORM_PRIORITY);

      System.out.println("End of totalTimeToWait.\n");

      outputFile.close();      // Close results file.
      ds.close();              // Close datagram socket.

      // The receiver should wait for a Stop Session to be received during 2DAYSTIMEOUT (Two days timeout is
      // a good value to prevent problems if we have two test sessions controled by the same OWAMP Control
      // session and one is much longer than the other). If a SS is not received all results should be invalidated.
      // Note: A Stop Session must be received by the receiver because it needs the information of Session
      // Description Record stored in the StopSession packet.

      switch(testStopSession(dataSPU.timeStamp.timeInMillisecondsSince1970(),lowerSequenceNumber,TWODAYSTIMEOUT)){
        case 0:// If a SS is not received the testStopSession will return 0. In this case all results should be invalidated.
          System.err.println("Error: Timeout waiting for the Stop Session packet with the Session Description Record information.");
          invalidateResults(true);  // invalidateResults(boolean sendSStoServer)
          return false;
        case 1:  // SS received
          break;
        case -1: // Error receiving SS
          return false;
        default:
          return false;
      }

      closeSocket(socketToCS); // close socket connection to Server

    }
    catch(IOException ex){
      System.err.println(ex);
      ds.close();              // Close datagram socket.
      closeSocket(socketToCS); // close socket connection to Server
      // The priority of the receiver Thread should be set to normal again.
      this.currentThread().setPriority(this.NORM_PRIORITY);
      return false;
    }

    return true;

  }


  /**
   * This method is used to test if a StopSession was sent by the Server.
   * @param lastSPDUtime double - represents the time of last SPDU received.
   * @param lowerSequenceNumber long - represents the lower sequence number received.
   * @param timeToWait int - time duration to wait for a Stop-Session.
   * @return short - It returns -1 if a failure occures or a SS with accept = 1 is received, 0 if none SS is received
   * and 1 if a SS with accept = 0 is received.
   */

  public short testStopSession(double lastSPDUtime, long lowerSequenceNumber, int timeToWait){

    // During timeToWait wait to receive StopSession from Server

    short stopVal = dataSSr.receiveSS(socketToCS, timeToWait);
    double stopSessionTime = hiResTimer.milliTime();  // Record the time of reception of the SS packet.

    if (stopVal == -2){ // The socket connections was reset.
      invalidateResults(false);  // invalidateResults(boolean sendSStoServer)
      return -1;
    }

    if ((stopVal == -1) || ( (stopVal == 1) && (dataSSr.accept != 0))) {
      if ( stopVal != -1) {
        System.out.println("StopSession from Server on " + socketToCS);
        verifyAcceptValue(dataSSr.accept, "Server",socketToCS, "");
        System.out.println();
        invalidateResults(false);  // invalidateResults(boolean sendSStoServer)
      }else{
        System.err.println("Error: Error receiving packet StopSession from Server in operation Stop Sessions.");
        System.out.println();
        invalidateResults(true);  // invalidateResults(boolean sendSStoServer)
      }

      return -1;
    }

    if (stopVal == 0) // None StopSessions were received
      return 0;

    // If a Stop Sessions was received and accept = 0, the receiver should receive the complete SS
    // and save the SDR in a file.

    System.out.println("Stop Sessions from Server on " + socketToCS);
    System.out.println();

    // Receive the Session Description Records and record them in the arrayOfSessionDescriptionRecords.
    arrayOfdataSDRr = new SessionDescriptionRecordReceiver[1];
    arrayOfdataSDRr = dataSSr.receiveCompleteSS(socketToCS);

    if(arrayOfdataSDRr == null){
      System.err.println("Error: Error in StopSession operation from Server.");
      System.err.println();
      // Invalidate this session results and send a SS with accept = 1 as the SS response to the Server.
      invalidateResults(true);  // invalidateResults(boolean sendSStoServer)
      return -1;
    }

    // The packet received was a Stop Session Packet and accept is zero

    /* If a receiver of an OWAMP-Test session learns through OWAMP-Control
     * Stop-Sessions message that the OWAMP-Test sender's last sequence
     * number is lower than any sequence number actually received, the
     * results of the complete OWAMP-Test session MUST be invalidated.
     */

    if ( (arrayOfdataSDRr[0].nextSeqno - 1) < lowerSequenceNumber) {

      System.err.println("Error: The OWAMP-Test sender's last sequence number is lower than any sequence number actually received");
      System.err.println();
      // Invalidate this session results and send a SS with accept = 1 as the SS response to the Server.
      invalidateResults(true);  // invalidateResults(boolean sendSStoServer)
      return -1;

    }

    /*  A receiver of an OWAMP-Test session, upon receipt of an OWAMP-Control
     * Stop-Sessions command, MUST discard any packet records -- including
     * lost packet records -- with a (computed) send time that falls between
     * the current time minus timeout and the current time. This ensures
     * statistical consistency for the measurement of loss and duplicates in
     * the event that the timeout is greater than the time it takes for the
     * Stop-Sessions command to take place.
     */

    try {
      int testPacketRecordLength = OwampTestPacketRecord.PACKET_LENGTH;
      String strSid = owampHexEncode(dataRS.sid);
      RandomAccessFile file = new RandomAccessFile(receiverTestResults.
                                                   getPath() +
                                                   receiverTestResults.separator +
                                                   strSid + ".owp", "rw"); // Open the results file for random access.

      // Verify timestamp of last packet received.

      if (lastSPDUtime + dataRS.timeout.timeInMillisecondsSince1900() >=
          stopSessionTime) {
        file.seek(file.length() - testPacketRecordLength);
        file.setLength(file.length() - testPacketRecordLength); // Set the new file length,
      } // to discard the last packet received.

      file.seek(8); // when the receiver receives a StopSession it should record the totalTimeTowait
      file.writeDouble(0); // with zero value because in a fetch session this value will be used to see if
      file.close();        // the session is still operating.
    }
    catch (IOException ex) {
      System.err.println(ex);
      System.err.println();
      // Invalidate this session results and send a SS with accept = 1 as the SS response to the Server.
      invalidateResults(true); // invalidateResults(boolean sendSStoServer)
      return -1;
    }

    // record the SDR in SDR file for this session.

    String strSid = owampHexEncode(arrayOfdataSDRr[0].sid);

    File sdrFileName = new File(receiverTestResults.getPath() +
                                 receiverTestResults.separator + strSid +
                                 ".sdr");


    if (sdrFileName.exists()){
      System.err.println("Error: The Session Description Record file already exists.");
      System.err.println();
      // Invalidate this session results and send a SS with accept = 1 as the SS response to the Server.
      invalidateResults(true); // invalidateResults(boolean sendSStoServer)
      return -1;
    }

    try{
      RandomAccessFile sdrFile = new RandomAccessFile(sdrFileName, "rw");

      // We just should record the Next Seqno and Number of Skip Ranges of the SDR, and the Skip Ranges in
      // the sdrFile.

      sdrFile.writeInt(arrayOfdataSDRr[0].nextSeqno);
      sdrFile.writeInt(arrayOfdataSDRr[0].numberOfSkipRanges);

      for (int i = 0; i < arrayOfdataSDRr[0].numberOfSkipRanges; i++) {
        sdrFile.writeInt(arrayOfdataSDRr[0].skipRangeDescriptions[i][0]);
        sdrFile.writeInt(arrayOfdataSDRr[0].skipRangeDescriptions[i][1]);
      }

      sdrFile.close();

    }catch (IOException ex) {
      System.err.println(ex);
      System.err.println();
      // Invalidate this session results and send a SS with accept = 1 as the SS response to the Server.
      invalidateResults(true); // invalidateResults(boolean sendSStoServer)
      return -1;
    }

    // If everything goes ok the receiver should Send the SS response to Server with accept = 0.

    System.out.println("Stop Session response to Server on "+socketToCS);
    System.out.println();

    dataSSs = new StopSessionsToBeSend( (short) 0, 0);
    dataSSs.sendSS(socketToCS);

    closeSocket(socketToCS);

    if (!ds.isClosed())
      ds.close();

    return 1;

  }


  /**
   * This method is used in method testStopSession to invalidate all the test results if an error ocurres
   * during the StopSession operation.
   * @param sendSStoServer boolean - used to verify if the receiver should send a SS to Server with accept = 1.
   */

  public void invalidateResults(boolean sendSStoServer){
    System.err.println(
        "The results of this OWAMP-Test session will be invalidate.");
    System.err.println();

    if (sendSStoServer) {
      System.out.println("StopSessions to Server on " + socketToCS);
      System.out.println();
      dataSSs = new StopSessionsToBeSend( (short) 1, 0);
      dataSSs.sendSS(socketToCS);
    }

    closeSocket(socketToCS); // Receiver should close connection to Server and stop receiving test packets.

    // The results of this session should be considered invalid. To do this, we delete the file of
    // packet records.

    String strSid = owampHexEncode(dataRS.sid);

    File resultsFileName = new File(receiverTestResults.getPath()+receiverTestResults.separator+strSid+".owp");
    File sdrFileName = new File(receiverTestResults.getPath()+receiverTestResults.separator+strSid+".sdr");

    if(resultsFileName.exists())
      resultsFileName.delete();

    if(sdrFileName.exists())
      sdrFileName.delete();

    if (!ds.isClosed())
      ds.close();

  }


  /**
   * Method to process de Fetch Session request.
   * The session receiver will respond with a Fetch-Ack message. Zero in the
   * Accept field means that the Session-Receiver is willing to send the session
   * results.  A value of 1 indicates rejection of the request.  All other values
   * are reserved.
   * @return boolean - This method returns true if the request operation was ok, otherwise it returns false.
   */

  public boolean processFetchSession(){

    int readn = receivePacket(socketToCS,buffer,1,FetchSession.PACKET_LENGTH -1);   // number of bytes read

    if(readn == (FetchSession.PACKET_LENGTH -1)){   // Packet Fetch Session

      System.out.println("Fetch Session from Server on "+socketToCS);
      System.out.println();

      // In variable byte[] buffer we have the packet FetchSession received

      FetchSessionReceiver dataFS = new FetchSessionReceiver();
      System.arraycopy(buffer,0,dataFS.buffer,0,dataFS.PACKET_LENGTH);
      if(!dataFS.receiveFS()){
        System.err.println("Error: Error in receiving FetchSession packet.\n");
        closeSocket(socketToCS);
        return false;
      }

      try{

        // If a complete session is requested and the session is still in
        // progress, or has terminated in any way other than normal, the request
        // to fetch session results MUST be denied. If an incomplete session is
        // requested, all packets received so far that fall into the requested
        // range SHOULD be returned.

        String strSid = owampHexEncode(dataFS.sid);

        File resultsFileName = new File(receiverTestResults.getPath()+receiverTestResults.separator+strSid+".owp");
        File sdrFileName = new File(receiverTestResults.getPath()+receiverTestResults.separator+strSid+".sdr");

        if(resultsFileName.exists()){

          if((dataFS.endSeq == 0xFFFFFFFF)&&(dataFS.beginSeq != 0x0)){  // If endSeq=-1 and beginSeq!=0 refuse FS.
            Fetch_AckToBeSend dataFA = new Fetch_AckToBeSend((byte)1);
            dataFA.sendFA(socketToCS);
            System.out.println("Fetch-Ack to Server on "+socketToCS);
            System.err.println("\nError: The Receiver didn't accept the command Fetch-Session request (endSeq = -1 and beginSeq != 0).");
            System.out.println();
            return false;
          }

                            // Test if the SessionReceiver file exists because when a receiver receives a not
                            // valid Stop Session packet or a StopSession packet with accept = 1, it must
                            // invalidate the results. To do this, the receiver deletes the packet record file.

          RandomAccessFile inFile = new RandomAccessFile(resultsFileName,"r");
          RandomAccessFile sdrFile = null;  // If the session is still operating we can make FS but no skip range descriptios are sent, because we don't have them yet.

          // Read StartTime and totalTimeToWait values from file

          double startTime = inFile.readDouble(), totalTimeToWait = inFile.readDouble();
          long filePointer = inFile.getFilePointer();  // Pointer to beginning of PacketRecord information
                                                       // in file.

          // Test if the session is still in progress. If yes the Server should reject
          // the request of FetchSession (we can not fetch the complete results of a session
          // that is still operating).

          if (dataFS.endSeq == 0xFFFFFFFF){   // a complete session is requested
            if((startTime+totalTimeToWait-1000) > hiResTimer.milliTime()){  // -1000 to prevent errors in case
              Fetch_AckToBeSend dataFA = new Fetch_AckToBeSend((byte)1);    // of a early FS.
              dataFA.sendFA(socketToCS);
              System.out.println("Fetch-Ack to Server on "+socketToCS);
              System.err.println("\nError: The Receiver didn't accept the command Fetch-Session request because");
              System.err.println("a complete Fetch-Session was requested before the end of the test session.");
              System.err.println("Possible cause of error: The systems are not syncronized.");
              System.out.println();
              inFile.close();
              closeSocket(socketToCS);
              return true;
            }
          }

          // Determine the numberOfPacketRecords stored in file

          int numberOfPacketRecords = 0;  // Number of packet records in file.
          numberOfPacketRecords=(int)((inFile.length()-16)/OwampTestPacketRecord.PACKET_LENGTH);

          // Determine the requested interval of packetRecord and the next seqno for this session.

          int firstPacket;
          int lastPacket;
          byte accept = 0;        // Used to determine if the Session-Receiver accepts the FS.
          byte finished = 1;      // Finished is non-zero if the OWAMP-Test session has terminated.

          if ( (startTime + totalTimeToWait - 1000) > hiResTimer.milliTime()) // OWAMP-Test session has not terminated yet. -1000 to prevent errors in case
            finished = 0;                                                     // of a early FS.
          else{
            // OWAMP-Test session has terminated. In this case we should verify if SS for this session
            // is already processed succesfully. This is done because at the end of a session a SS is sent by the CC
            // with the session skip range information. This information is used in the Fecth-Session.
            // When SS is complete the skip range information will be stored in the sdrFile. So if sdrFile doesn't
            // exists it means SS doesn't finish yet. But we should wait only for TIMEOUT seconds.
            int time=0;
            while(!sdrFileName.exists()&&(time<TIMEOUT)){
              time += 10000;
              try{
                sleep(10000);
              }catch(Exception e){}
            }
            if(!sdrFileName.exists()){   // In this case the receiver should reject the request of FetchSession.
              Fetch_AckToBeSend dataFA = new Fetch_AckToBeSend((byte)1);
              dataFA.sendFA(socketToCS);
              System.out.println("Fetch-Ack to Server on "+socketToCS);
              System.err.println("\nError: The Receiver didn't accept the command Fetch-Session request.");
              System.err.println("Error: There are no results for this session (The Session Description Records file is missing).");
              System.out.println();
              inFile.close();
              closeSocket(socketToCS);
              return true;
            }
            sdrFile = new RandomAccessFile(sdrFileName,"r");
            finished = 1;
          }

          int nextSeqno = 0;      // Next Seqno indicates the next sequence number that would have been
                                  // sent from this send session.

          int numSkipRanges = 0;  // Number of Skip Ranges indicates the number of holes that actually
                                  // occurred in the sending process.

          if(finished != 0){  // OWAMP-Test session has terminated, read the nextSeqno and numSkipRanges from sdrFile.
            nextSeqno = sdrFile.readInt();
            numSkipRanges = sdrFile.readInt();
          }

          int numPacketRecords = 0;   // the number of Packet Records in file that belongs to the interval
                                      // of packet records requested.
          int lastPacketAux = 0;      // This variable is used to identify the maximum sequence number found
                                      // in the requested interval, in case of a incomplite FS.

          OwampTestPacketRecordToBeSend packetRecord = new OwampTestPacketRecordToBeSend();
          int i = 0; // Used to control for iteration;

          if ((dataFS.beginSeq == 0x0)&&(dataFS.endSeq == 0xFFFFFFFF)) {  // a complete session is requested
            firstPacket = 0;
            lastPacket = numberOfPacketRecords;
            numPacketRecords = numberOfPacketRecords;
          }else {  // only a part of the session is requested
            firstPacket = dataFS.beginSeq;
            lastPacket = dataFS.endSeq;

            // Verify if it exists any packet record requested in the file of packet records and, if the
            // OWAMP-Test session has terminated, determine the next seqno for this session.

            for (i = 0; i < numberOfPacketRecords; i++) {
              packetRecord.sequenceNumber=inFile.readInt();
              if((packetRecord.sequenceNumber >= firstPacket)&&(packetRecord.sequenceNumber <= lastPacket)){
                numPacketRecords ++;
                lastPacketAux = (packetRecord.sequenceNumber >= lastPacketAux)?packetRecord.sequenceNumber:lastPacketAux;
              }

              inFile.skipBytes(packetRecord.PACKET_LENGTH - 4);
            }

          }

          // Send the Fetch-Ack packet and the interval of PacketRecords requested, if they exits.

          Fetch_AckToBeSend dataFA = new Fetch_AckToBeSend(accept, finished,
                nextSeqno, numSkipRanges, numPacketRecords);
          dataFA.sendFA(socketToCS);
          System.out.println("Fetch-Ack to Server on " + socketToCS);
          System.out.println();

          // Send the OWAMP Test Session Data to the server
          // The OWAMP-Test session data consists of the following (concatenated):


          // I - A reproduction of the Request-Session command that was used to start the session; it is modified
          // so that actual sender and receiver port numbers that were used by the OWAMP-Test session always appear
          // in the reproduction (This will be done by the Server).

          System.out.println("OWAMP Test Session Data to the server on "+socketToCS);
          System.out.println();

          // II - Zero or more (as specified) Skip Range descriptions.

          Packets packet = null;
          if(numSkipRanges != 0)
            packet = new Packets(numSkipRanges*8);

          for (i = 0; i < numSkipRanges;i++){
            packet.Int_To_Buffer(sdrFile.readInt(), i*8);    // firstSeqnoSkipped
            packet.Int_To_Buffer(sdrFile.readInt(), 4+i*8);  // lastSeqnoSkipped
          }
          if((finished != 0)&&sdrFileName.exists())
            sdrFile.close();

          if(numSkipRanges != 0){
            packet.sendTCP(socketToCS,numSkipRanges*8);

            // Each Skip Range Descriptions has only 8 bytes, so the last (possibly full, possibly incomplete)
            // block (16 octets) of data MUST be padded with zeros, if necessary (numberOfSkipRanges is odd).
            // This ensures that the next session description record starts on a block boundary.
            if ((numSkipRanges % 2) == 1) {
              // Send the last 8 bytes of zeros to complete the 16 octets of the last block.
              packet.Int_To_Buffer(0,0);
              packet.Int_To_Buffer(0,4);
              packet.sendTCP(socketToCS,8);
            }
          }

          // III - 16 octets of IZP.

          IntegrityZeroPaddingToBeSend izp = new IntegrityZeroPaddingToBeSend(16);   // IZP of 16 bytes
          izp.sendIZP(socketToCS);  // Send the IZP

          // IV - Zero or more (as specified) packet records.

          // Send the requested packet records

          for (i = firstPacket; i <= lastPacketAux;i++){
            inFile.seek(filePointer);   // Put the file pointer in start of file.
            for (int c = 0; c < numberOfPacketRecords; c++){
              packetRecord.sequenceNumber=inFile.readInt();
              // If packetRecord.sequenceNumber == i send the packetRecord. Or if a complete session is
              // requested send all packetRecord.
              if ((packetRecord.sequenceNumber == i) || (dataFS.endSeq == 0xFFFFFFFF)) {
                packetRecord.sendTimestamp.integerPartOfSeconds=inFile.readInt();
                packetRecord.sendTimestamp.fractionalPartOfSeconds=inFile.readInt();
                packetRecord.sendError.s_z_scale=inFile.readByte();
                packetRecord.sendError.multiplier=inFile.readByte();
                packetRecord.receiveTimestamp.integerPartOfSeconds=inFile.readInt();
                packetRecord.receiveTimestamp.fractionalPartOfSeconds=inFile.readInt();
                packetRecord.receiveError.s_z_scale=inFile.readByte();
                packetRecord.receiveError.multiplier=inFile.readByte();
                packetRecord.ttl = inFile.readByte();
                          // Send packet record
                packetRecord.sendOTPR(socketToCS);

              }else
                inFile.skipBytes(packetRecord.PACKET_LENGTH - 4);

              // Sleep for sleepVal ms before sending the next packet. We do this to
              // prevent overload of the Receiver processor PC.
              sleepTime(sleepVal);

            }

            if(dataFS.endSeq == 0xFFFFFFFF)   // a complete session is requested. All packetRecord are sent
              break;                          // for i = 0, we should break here.
          }


          // V - 16 octets of IZP.

          izp.sendIZP(socketToCS);  // Send the IZP

          inFile.close();

          if (deleteFetchedResults && (dataFS.endSeq == 0xFFFFFFFF)){   // If a complete session is requested, delete file at the end.
            resultsFileName.delete();
            sdrFileName.delete();
            File RSfileName = new File(receiverRSdir.getPath()+receiverRSdir.separator+strSid+".rs");
            RSfileName.delete();
          }

        }else{  // In this case the receiver should reject the request of FetchSession.

          Fetch_AckToBeSend dataFA = new Fetch_AckToBeSend((byte)1);
          dataFA.sendFA(socketToCS);
          System.out.println("Fetch-Ack to Server on "+socketToCS);
          System.err.println("\nError: The Receiver didn't accept the command Fetch-Session request.");
          System.err.println("Error: There are no results for this session (The results file is missing).");
          System.err.println();
          closeSocket(socketToCS);
          return true;

        }

      }catch(IOException ex){
        ex.printStackTrace();
        closeSocket(socketToCS);
        return false;
      }

      closeSocket(socketToCS);
      return true;

    }else
      System.err.println("Error: The packet received in socket "+socketToCS+" is not a Fetch Session packet.\n");

    // If we reach this point it means that the packet received doesn't respects the protocol or an error occurred.
    // So, we should stop this OWAMP-Control session.

    if(readn > 0)   // Packet received has a valid type but not a valid length.
      System.err.println("Error: The packet received doesn't exits in the protocol.\n");

    if(readn == 0)  // Timeout to wait occurred.
      System.err.println("Error: Timeout to wait for the packet occurred.\n");
    System.err.println("This session will be finished.");
    System.err.println();

    closeSocket(socketToCS);

    return false;

  }


  /**
   * This method is used to stop the processing and wait the time in ms specified by variable timeToWait.
   * It is used to wait for StartTime. During this interval the Receiver should test if the Server didn't
   * send a Stop session command.
   * @param timeToWait long - the time to wait before continuing.
   * @return short - It returns -1 if a failure occures or a SS with accept = 1 is received, 0 if none SS is received
   * and 1 if a SS with accept = 0 is received.
   */

  public short waitTimeToWait(double timeToWait){

    if(timeToWait <= 0)
      timeToWait = 1;

    // During timeToWait wait to receive StopSession from Server

    int stopVal = dataSSr.receiveSS(socketToCS,timeToWait);

    if((stopVal == -2)||(stopVal == -1)){  // Socket reset or an error occurred.
      if(stopVal == -1)
        System.err.println("Error: Error receiving packet StopSession from Server in operation Stop Sessions.");
      System.err.println("Error: This Session will not be started.\n");
      closeSocket(socketToCS);  // Receiver should close connection to Server and stop receiving test packets.
      if(!ds.isClosed())
        ds.close();
      return -1;
    }

    if(stopVal == 0)  // None StopSessions were received
      return 0;

    // If a Stop Session is recieved the Receiver should close connection
    // to Server and doesn't start the receiving process.

    System.out.println("StopSession from Server on " + socketToCS);

    if(dataSSr.accept == 0){
      // Receive the Session Description Records and record them in the arrayOfSessionDescriptionRecords.
      arrayOfdataSDRr = new SessionDescriptionRecordReceiver[1];
      arrayOfdataSDRr = dataSSr.receiveCompleteSS(socketToCS);

      if (arrayOfdataSDRr == null)
        System.err.println("Error: Error in StopSession operation from Server.");
    }

    if(!verifyAcceptValue(dataSSr.accept, "Server",socketToCS, "")){
      // If accept == 0 (premature SS) the receiver should Send the SS response to Server with accept = 0.
      dataSSs = new StopSessionsToBeSend( (short) 0, 0);
      dataSSs.sendSS(socketToCS);
    }

    System.err.println("Error: This Session will not be started.");
    System.err.println();

    closeSocket(socketToCS);
    if(!ds.isClosed())
      ds.close();

    return 1;

  }


  // Main function

  public static void main(String[] args) throws IOException{

    InetAddress receiverAddress = null; // The local IP address to bind to. The receiverAddress can be used on a
                                        // multi-homed host for a ServerSocket that will only accept connect requests
                                        // to one of its addresses. If serverAddress is null, it will default
                                        // accepting connections on any/all local addresses.
    int port = -1,            // TCP port number to be used in OWAMP-Control.
        sleepVal = -1;        // Used to Sleep for sleepVal ms before sending the next packet record packet in
                              // Fetch-Session operation. We do this to prevent overload of the Receiver processor PC.
    int[] portRange = null;   // Represents the UDP port range to be used during test sessions.
    boolean keepResults = false;     // Used to verify if the configuration and the results of a session should be deleted after
                                     // a Fetch-Session of all results is completed.
    String  ntpServerAddress = null; // The address of the NTP server to be used to get the time reference.
                                     // If not defined the system time will be used to get the time reference.
    int updateInterval = -1;         // The interval duration (in minutes) between two consecutives updates to time reference.

    if((args.length == 1)&&((args[0].compareToIgnoreCase("-?") == 0)||(args[0].compareToIgnoreCase("-help") == 0))){  // Print the help message
      System.out.println("\nUsage: OWAMP_SessionReceiver [-options]\n");
      System.out.println("where options include:");
      System.out.println("\n   -a | -address                     \tThe local IP address to bind to. This address can be used on a multi-homed host");
      System.out.println("                                       \tfor a Session-Receiver that will only accept connection requests to one of its addresses.");
      System.out.println("                                       \tIf not defined, it will by default accept connections on any/all local addresses.");
      System.out.println("\n   -p | -port port                   \tThe TCP port number to be used in OWAMP-Control.");
      System.out.println("\n   -r | -portRange [lowport,highport]\tThe UDP port range to be used in OWAMP-Test sessions.");
      System.out.println("\n   -s | -sleepVal sleepVal           \tThe number of milliseconds to wait between packet record sending");
      System.out.println("                                       \tprocess on Fetch Sessions. This value is used to prevent overloading");
      System.out.println("                                       \tthe Session Receiver's machine CPU during Fetch Sessions.");
      System.out.println("\n   -k | -keepResults                 \tIf defined the configuration and the results of the test session will be keeped,");
      System.out.println("                                       \tafter the processing of a Fetch-Session requesting all the results of that");
      System.out.println("                                       \ttest session. Otherwise the configuration and the results are deleted (Used by default).");
      System.out.println("\n   -n | -ntpServer                   \tThe address of the NTP server to be used to get the time reference. If not defined,");
      System.out.println("                                       \tthe system time will be used to get the time reference.");
      System.out.println("\n   -u | -updateInterval              \tThe value (in minutes) to be used for the time interval between two consecutives");
      System.out.println("                                       \tupdates of the time reference.");
      System.out.println("\n   -? | -help                        \tPrints this help message.");
      System.exit(1);
    }

    if(args.length == 0){
      System.out.println("\nDefault ports and sleepVal will be used.");
      System.out.println("\nThe configuration and results of each session will be deleted after the processing of a Fetch-Session requesting all the results of the session.");
      port = DEFAULT_RECEIVER_PORT;
      sleepVal = 10;
      keepResults = false;
    }else{
      try{
       int numOfArg = 0; // Used to determine if all arguments are ok.

       // Note: In this case we should verify each argument because the -k option only occupies one argument
       // and not two. So we should use i++ instead of i+=2 in the for cicle. This is done because the -k
       // option can be in any argument.

       // option -address (It's the local address to bind to)
       for (int i = 0; i < args.length; i++) {
         if ((args[i].compareToIgnoreCase("-a") == 0)||(args[i].compareToIgnoreCase("-address") == 0)) {
           try{ // Test if the receiver address introduced is a valid address.
             receiverAddress = InetAddress.getByName(args[i + 1]);
             if(!OWAMP_Base.isAnyLocalAddress(receiverAddress)){
               System.err.println("Error: The OWAMP Session-Receiver's address read is not a valid local address (" + args[i + 1] + ").");
               System.exit(1);
             }
           }catch (Exception e) {
             System.err.println(
                 "Error: The OWAMP Session-Receiver's address read is not a valid address (" +
                 args[i + 1] + ").");
             System.exit(1);
           }
           numOfArg+=2;
           break;
         }
       }

       // option -port
       for (int i = 0; i < args.length; i++) {
         if ((args[i].compareToIgnoreCase("-p") == 0)||(args[i].compareToIgnoreCase("-port") == 0)) {
           port = Integer.parseInt(args[i + 1]);

           if (port < 0 || port >= 65536) {
             System.err.println("Error: Receiver TCP port number out of range (" +
                                port + ").");
             System.err.println();
             System.exit(1);
           }
           numOfArg+=2;
           break;
         }
       }

       // option -portRange
       for (int i = 0; i < args.length; i++) {

         if ((args[i].compareToIgnoreCase("-r") == 0)||(args[i].compareToIgnoreCase("-portRange") == 0)) {
           int upperIndex = args[i + 1].indexOf(",");
           int lowRange = Integer.parseInt(args[i + 1].substring(1, upperIndex)),
               upperRange = Integer.parseInt(args[i + 1].substring(upperIndex + 1,args[i + 1].indexOf("]")));

           if (lowRange < 0 || lowRange >= 65536) {
             System.err.println("Error: Lower range TCP port number out of range (" +
                                lowRange + ").");
             System.err.println();
             System.exit(1);
           }
           if (upperRange < 0 || upperRange >= 65536) {
             System.err.println("Error: Upper range TCP port number out of range (" +
                                upperRange + ").");
             System.err.println();
             System.exit(1);
           }

           if(upperRange - lowRange < 0){
             System.err.println("\nError: Upper port range should be greather than lower port range.");
             throw new IllegalArgumentException();
           }

           portRange = new int[2];
           portRange[0] = lowRange;
           portRange[1] = upperRange;
           numOfArg+=2;
           break;
         }
       }

       // option -sleepVal
       for (int i = 0; i < args.length; i++) {
         if ((args[i].compareToIgnoreCase("-s") == 0)||(args[i].compareToIgnoreCase("-sleepVal") == 0)) {
           sleepVal = Integer.parseInt(args[i + 1]);
           if (sleepVal < 0) {
             System.err.println("Error: The value of sleepVal can not be a negative number (" +
                                sleepVal + ").");
             System.err.println();
             System.exit(1);
           }
           numOfArg+=2;
           break;
         }
       }

       // option -keepResults
       for (int i = 0; i < args.length; i++) {
         if ((args[i].compareToIgnoreCase("-k") == 0)||(args[i].compareToIgnoreCase("-keepResults") == 0)) {
           keepResults = true;
           numOfArg++;
           break;
         }
       }

       // option -ntpServer
       for (int i = 0; i < args.length; i++) {
         if ((args[i].compareToIgnoreCase("-n") == 0)||(args[i].compareToIgnoreCase("-ntpServer") == 0)) {
           ntpServerAddress = args[i + 1];
           try{ // Test if the NTP server address introduced is a valid address.
             if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
               System.err.println(
                   "Error: The NTP server's address read is not a valid address (" +
                   InetAddress.getByName(ntpServerAddress) + ").");
               System.exit(1);
              }
           }catch (Exception e) {
             System.err.println(
                 "Error: The NTP server's address read is not a valid address (" +
                 ntpServerAddress + ").");
             System.exit(1);
           }
           numOfArg+=2;
           break;
         }
       }

       // option -updateInterval
       for (int i = 0; i < args.length; i++) {
         if ((args[i].compareToIgnoreCase("-u") == 0)||(args[i].compareToIgnoreCase("-updateInterval") == 0)) {
           updateInterval = Integer.parseInt(args[i + 1]);
           if (updateInterval < 0) {
             System.err.println("Error: The value of updateInterval can not be a negative number (" +
                                updateInterval + ").");
             System.err.println();
             System.exit(1);
           }
           numOfArg+=2;
           break;
         }
       }

       if((numOfArg<args.length)||(receiverAddress==null)&&(port==-1)&&(portRange==null)&&(sleepVal==-1)&&(!keepResults)&&(ntpServerAddress==null)&&(updateInterval==-1))
          throw new IllegalArgumentException();

      }catch (Exception e){
        System.err.println("\nIllegalArgumentException in thread main: Usage: OWAMP_SessionReceiver [-options]\n");
        System.err.println("where options include:");
        System.err.println("\n   -a | -address                     \tThe local IP address to bind to. This address can be used on a multi-homed host");
        System.err.println("                                       \tfor a Session-Receiver that will only accept connection requests to one of its addresses.");
        System.err.println("                                       \tIf not defined, it will by default accept connections on any/all local addresses.");
        System.err.println("\n   -p | -port port                   \tThe TCP port number to be used in OWAMP-Control.");
        System.err.println("\n   -r | -portRange [lowport,highport]\tThe UDP port range to be used in OWAMP-Test sessions.");
        System.err.println("\n   -s | -sleepVal sleepVal           \tThe number of milliseconds to wait between packet record sending");
        System.err.println("                                       \tprocess on Fetch Sessions. This value is used to prevent overloading");
        System.err.println("                                       \tthe Session Receiver's machine CPU during Fetch Sessions.");
        System.err.println("\n   -k | -keepResults                 \tIf defined the configuration and the results of the test session will be keeped, after");
        System.err.println("                                       \tthe processing of a Fetch-Session requesting all the results of that");
        System.err.println("                                       \ttest session. Otherwise the configuration and results are deleted (Used by default).");
        System.err.println("\n   -n | -ntpServer                   \tThe address of the NTP server to be used to get the time reference. If not defined,");
        System.err.println("                                       \tthe system time will be used to get the time reference.");
        System.err.println("\n   -u | -updateInterval              \tThe value (in minutes) to be used for the time interval between two consecutives");
        System.err.println("                                       \tupdates of the time reference.");
        System.err.println("\n   -? | -help                        \tPrints this help message.");
        System.exit(1);
      }

      if(port == -1){
        port = DEFAULT_RECEIVER_PORT;
        System.out.println("\nDefault port for OWAMP-Control will be used ("+DEFAULT_RECEIVER_PORT+").");
      }

      if(portRange==null){
        System.out.println("\nDefault portRange for OWAMP-Test will be used.");
      }

      if(sleepVal == -1){
        sleepVal = 10;
        System.out.println("\nDefault sleepVal will be used ("+sleepVal+" ms).");
      }

      if(!keepResults){
        System.out.println("\nThe configuration and the results of each session will be deleted after the processing of a Fetch-Session requesting all the results of the session.");
      }

    }

    System.out.println("\nStarting on port "+port);
    System.out.println();

    try{
      ServerSocket receiverSocket = null;
      if(receiverAddress == null)
        receiverSocket = new ServerSocket(port); // initialize a server socket on port 'port'
      else
        receiverSocket = new ServerSocket(port,0,receiverAddress); // initialize a server socket on port 'port'

      // The hiResTimer should be definied here to guarante that
      // all Receiver instantiation has the same time reference. It will
      // be passed has an argument of the Receiver constructor.
      HiResTimer hiResTimer = null;
      if(updateInterval == -1){  // Use default updateInterval defined on HiResTimer class.
        hiResTimer = new HiResTimer(ntpServerAddress);
      }else
        hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

      Timestamp upTime = new Timestamp(hiResTimer.milliTime(), (short)1);
                                                  // Uptime is a timestamp representing the time when the
                                                  // current instantiation of the Receiver started operating, in
                                                  // milliseconds since 1Jan1970.

      // Test receiverDis directory. If it exists it should be cleaned (delete it and create it again).
      // This directory will have all RS (in receiverRSdir directory) and results (in receiverTestResults directory)
      // received from senders in this Receiver.

      if(OWAMP_SessionReceiver.receiverDir.exists())
        try{
          OWAMP_SessionReceiver.deleteDir(OWAMP_SessionReceiver.receiverDir);
        }
        catch(IOException ex){
          ex.printStackTrace();
        }

      OWAMP_SessionReceiver.receiverDir.mkdir();
      OWAMP_SessionReceiver.receiverRSdir.mkdir();
      OWAMP_SessionReceiver.receiverTestResults.mkdir();

      // wait for requests connections

      while(true){

        Socket server = receiverSocket.accept();    // accept the server socket connection. It can be a socket
                                                    // connection to process a Test Session or a Fetch Session.

        OWAMP_SessionReceiver receiver = new OWAMP_SessionReceiver(server, portRange, upTime, sleepVal, keepResults, hiResTimer); // deliver this socket to the receiver

        receiver.mode = 1;                          // Mode of authentication.
        receiver.start();                           // initialize thread

      }
    }
    catch (SocketException se){
      System.err.println(se);
    }
  }

}
